"""Tests for string parsing."""

import pytest

from reno import ops, parser
from reno.components import Distribution, Flow, Piecewise, Scalar, Stock, Variable
from reno.model import Model


@pytest.mark.parametrize(
    "string,expected,expected_type",
    [
        ("5.0", 5.0, float),
        (" 5.0", 5.0, float),
        ("5.0 ", 5.0, float),
        (" 5.0 ", 5.0, float),
        ("13", 13, int),
        (" 13", 13, int),
        (" 13 ", 13, int),
        ("10", 10, int),
        ("something", "something", str),
    ],
)
def test_int_float_parsing(string, expected, expected_type):
    val = parser.parse_value(string)
    assert isinstance(val, expected_type)
    assert val == expected


@pytest.mark.parametrize(
    "string,expected_args,expected_kwargs",
    [
        ("Normal(5.0, 10)", [5.0, 10], {}),
        ("Normal(5.0, std=10)", [5.0], {"std": 10}),
        ("Normal(5.0, std =13.2)", [5.0], {"std": 13.2}),
        ("Scalar(1.0)", [1.0], {}),
        ("Scalar([1.0, 6.0, 2.0])", [[1.0, 6.0, 2.0]], {}),
        ("Normal(Scalar(5.0), Scalar(10.0))", [Scalar(5.0), Scalar(10.0)], {}),
        ("Normal(Scalar(5.0), 10.0)", [Scalar(5.0), 10.0], {}),
        ("Normal(Scalar(5.0), std=Scalar(10.0))", [Scalar(5.0)], {"std": Scalar(10.0)}),
        ("Normal(Scalar(5.0), 10.0, False)", [Scalar(5.0), 10.0, False], {}),
    ],
)
def test_param_parsing(string, expected_args, expected_kwargs):
    args, kwargs, *_ = parser.parse_function_args(string)
    for i, arg in enumerate(args):
        if isinstance(arg, Scalar):
            assert arg.value == expected_args[i].value
        else:
            assert arg == expected_args[i]
    for kw, arg in kwargs.items():
        if isinstance(arg, Scalar):
            assert arg.value == expected_kwargs[kw].value
        else:
            assert arg == expected_kwargs[kw]
    # assert args == expected_args
    # assert kwargs == expected_kwargs


def test_full_op_parse():
    out = parser.parse_class_or_scalar(" Normal(5.0, std =13.2)")
    assert isinstance(out, ops.Normal)
    assert out.sub_equation_parts[0].value == 5.0
    assert out.sub_equation_parts[1].value == 13.2


def test_full_assignment_op_parse():
    out = parser.parse(" (= (= Scalar(5.0)))")
    assert isinstance(out, ops.assign)


def test_blank_parse():
    assert parser.parse_class_or_scalar(" ") is None


@pytest.mark.parametrize(
    "string,expected_name,expected_args",
    [
        ("(+ 5 3)", "+", ["5", "3"]),
        ("(+ (5 3))", "+", ["(5 3)"]),
        ("(+ (5) 3)", "+", ["(5)", "3"]),
        ("(+ ((5)) 3)", "+", ["((5))", "3"]),
        ("(+ 'thing' 3)", "+", ["'thing'", "3"]),
        ("(+ (+ 2 'thing') 3)", "+", ["(+ 2 'thing')", "3"]),
        ("+ 5 3", "+", ["5", "3"]),
        ("+ 5 (+ 3 6)", "+", ["5", "(+ 3 6)"]),
        ("(+ 5 (+ 3 6))", "+", ["5", "(+ 3 6)"]),
        ("(= Scalar(5.0))", "=", ["Scalar(5.0)"]),
        ("(= (= Scalar(5.0)))", "=", ["(= Scalar(5.0))"]),
    ],
)
def test_prefix_op_str(string, expected_name, expected_args):
    """We should be able to separate out arguments in raw strings
    of prefix syntax."""
    name, args = parser.parse_op_str(string)

    assert name == name
    assert args == expected_args


@pytest.mark.parametrize(
    "string,expected_result",
    [
        ("(+ 5 3)", 8),
        ("(+ (- 5 2) 3)", 6),
        ("(+ (- 5 (maximum 3 4)) (minimum 2 1))", 2),
    ],
)
def test_prefix_parsing_math(string, expected_result):
    """Basic math operations should parse and evaluate correctly."""
    op = parser.parse(string)
    assert op.eval(0) == expected_result


def test_prefix_parsing_w_resolution():
    """References should correctly resolve during prefix parsing."""
    v = Variable(Scalar(2))
    v.name = "v"
    v.populate(1, 1)

    op = parser.parse("(+ 3 'v')", {"v": v})
    assert op.eval(0) == 5


def test_prefix_parsing_w_piecewise_example():
    m = Model()
    m.decision = Variable(ops.Bernoulli(0.5))
    m.f = Flow(
        Piecewise(
            [Scalar(100), Scalar(0)],
            [ops.eq(m.decision, Scalar(1)), ops.eq(m.decision, Scalar(0))],
        )
    )
    m.s = Stock()
    m.s += m.f

    p_op = parser.parse(str(m.f.eq), {r.qual_name(): r for r in m.all_refs()})
    assert len(p_op.conditions) == 2
    assert len(p_op.equations) == 2

    assert m.decision in p_op.conditions[0].seek_refs()


def test_prefix_mixed_classsyntax():
    """The distribution reprs are formatted like python functions, parsing
    should handle this appropriately."""
    op = parser.parse(str(ops.sum(ops.Bernoulli(0.5))))
    assert isinstance(op.sub_equation_parts[0], Distribution)


def test_parse_multiple_arrays():
    """An op with multiple array parameters shouldn't get mixed up."""
    m = Model()
    m.v = Variable()
    op_str = str(
        ops.interpolate(
            m.v,
            # Scalar([1.0, 2.0, 3.0]),
            # Scalar([2.0, 3.0, 4.0]))
            Scalar([0.0, 0.2, 0.4, 0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0]),
            Scalar([0.05, 0.105, 0.225, 0.36, 0.54, 0.84, 1.24, 2.36, 3.34, 3.86, 4.0]),
        )
    )
    print(op_str)
    op = parser.parse(op_str, refs={"v": m.v})

    assert isinstance(op, ops.interpolate)
    assert (
        op.sub_equation_parts[1].value
        == [0.0, 0.2, 0.4, 0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0]
    ).all()



"""Utility functions that are potentially needed in multiple modules."""

import importlib.resources
import inspect
from typing import Any

import numpy as np

import reno

# these are the default colors to use when rendering 'selected'
# references in latex. Might want to change in disgusting lightmode contexts
DOC_HIGHLIGHT_COLOR = "teal"
EQ_HIGHLIGHT_COLOR = "cyan"


def _get_assigned_var_name(var: Any) -> str:
    """In a model context manager we don't have the direct ability to get the name of the variable
    we're assigning to the model (to avoid the PyMC requirement of specifying the name separately
    as a string.)

    To address that, this function pulls from the local frame to find the variable matching the
    one passed.

    https://stackoverflow.com/questions/18425225/getting-the-name-of-a-variable-as-a-string
    """
    # NOTE: this can't be used within the __init__ of a reference because the
    # variable on the left hand side of the assignment isn't technically in the
    # frame yet/hasn't been created.
    frame_items = inspect.currentframe().f_back.f_back.f_locals.items()
    for var_name, var_val in frame_items:
        if var_val is var:
            return var_name

    # TODO: should this throw an error instead?
    return None


def find_ref_root_model(ref):
    """Find the top most parent model in the submodel tree."""
    model = ref.model
    while model.parent is not None:
        model = model.parent
    return model


def ref_universe(
    refs: list["reno.components.TrackedReference"], depth=1
) -> list["reno.components.TrackedReference"]:
    """Fancier recursive seek_refs, useful for collecting the full set of
    references to and from any in the given set of passed refs.

    Given a set of references, expand outwards ``depth`` times to find other
    references used.

    Args:
        refs (list[reno.components.TrackedReference]): The references to start expanding outwards from.
        depth (int): How many steps outward to take finding dependencies from the initial references list.

    Returns:
        A list (including the initial references) of expanded dependencies/dependents.
    """
    universe = [*refs]
    while depth > 0:
        new_refs = []
        for ref in universe:
            # check for "inputs"
            if isinstance(ref, reno.components.Stock):
                new_refs.extend(ref._implied_eq().seek_refs())
            else:
                new_refs.extend(ref.seek_refs())

            # check for "outputs"
            for ref2 in find_ref_root_model(ref).all_refs():
                if isinstance(ref2, reno.components.Stock):
                    if ref in ref2._implied_eq().seek_refs():
                        new_refs.append(ref2)
                else:
                    if ref in ref2.seek_refs():
                        new_refs.append(ref2)

        universe.extend(new_refs)
        universe = list(set(universe))
        depth -= 1

    return universe


def is_ref_in_parent_scope(ref, source) -> bool:
    """Check if a reference is in the scope _above_ some source reference.
    (In other words, in the source model's parent, or parent of that parent,
    up the chain, anything except the same model as source)

    This is primarily used in the diagramming logic to determine which
    model/submodel needs to handle rendering a particular reference.

    Args:
        ref (reno.components.TrackedReference): The reference to look for in
            parent scopes.
        source (reno.components.TrackedReference): The reference to search
            the parent chain through.

    Returns:
        ``True`` if a ref is somewhere in the source's parent chain
        (and notably _not_ in the same exact model)
    """
    if ref.model is None or source.model is None:
        # TODO: arguably if source's model is NOT None and ref's is,
        # that should count as rendering "external to the current model"
        return False

    parent = source.model.parent
    while parent is not None:
        if ref.model == parent:
            return True
        parent = parent.parent
    return False


def ensure_scalar(operand: Any) -> Any:
    """Convert int/float types into a Scalar type if relevant, otherwise
    directly return what was passed."""
    if isinstance(operand, (int, float, np.ndarray)):
        return reno.components.Scalar(operand)
    if isinstance(operand, list):
        return reno.components.Scalar(np.array(operand))
    return operand


def is_free_var(eq: "reno.components.EquationPart") -> bool:
    """A reference is a "free reference/free variable" if its equation
    has no references to _other_ references.

    Expects an EquationPart rather than only the ref itself because this applies
    both to stock .init as well as flow/var .eq
    """
    if (
        isinstance(
            eq,
            (
                reno.components.Scalar,
                reno.components.Distribution,
                int,
                float,
                np.ndarray,
            ),
        )
        or eq is None
        or (isinstance(eq, reno.ops.assign) and is_free_var(eq.sub_equation_parts[0]))
    ):
        return True
    return False


def check_for_easy_static_time_eq(eq: "reno.components.EquationPart") -> bool:
    """Determine whether an equation is `t - STATIC_EQ`.

    This is necessary for basic pymc optimizations on accessing historical values.
    (Basically any index equations not following this form are gonna be a whole lot
    harder, and I'm not going to worry about implementing them yet.)
    """

    if not isinstance(eq, reno.ops.sub):
        return False
    if not isinstance(eq.sub_equation_parts[0], reno.components.TimeRef):
        return False
    if (
        not is_static(eq.sub_equation_parts[1])
        or len(eq.sub_equation_parts[1].find_refs_of_type(reno.components.Distribution))
        > 0
    ):
        # NOTE: currently not counting distributions as "easy" because of
        # weirdness in computing the index multiple times resulting in multiple
        # values (thus pymc taps are wrong)
        return False
    return True


def is_static(
    eq: "reno.components.EquationPart",
    checked: list["reno.components.Reference"] = None,
) -> bool:
    """An equation is "static" if it only contains values that won't change, e.g. scalars,
    distributions, or other static variables. Specifically stock and time references are
    what make an equation _not_ static.

    Args:
        eq (EquationPart): The equation to determine static status of.
        checked (list[Reference]): Used internally to avoid infinite recursion in checking,
            don't manually pass.
    """
    # print("Called is_static on", eq)
    if eq is None:
        # really?? I guess this makes sense, none isn't going to change if it even runs.
        # print("No eq, so not static")
        return True
    if checked is None:
        checked = []
    # the reason we're throwing in ref checks instead of just checking
    # isinstance type is because you could theoretically have an eq of
    # e.g. Scalar(4) + Scalar(5) which is obviously still static
    for ref in eq.seek_refs():
        # print("Checking", ref)
        # the hasattr is to check for already non-tracked references (e.g. TimeRef)
        if not hasattr(ref, "_static"):
            # print("No _static, so not static")
            return False

        # TODO: TODO: TODO: WHAT A MESS. CLEAN THIS UP.

        # if ref.eq is not None:
        if ref not in checked or ref._static is None:
            checked.append(ref)
            if not is_static(ref, checked):
                return False
        else:
            # TODO: we don't set _static in here though, this would only
            # happen from TrackedReference._determine_if_static. (which is
            # similarly calling this is_static function)
            # Is the expectation that this condition is only for when that
            # sequence of _determine_if_static() calls during population?
            # It feels really weird to be using a piece of state for what
            # should likely be a stateless check?
            if not ref._static:
                # print("Ref's _static is false, so not static")
                return False

        # if ref.eq is not None and ref not in checked:
        #     # something something without checked we got inf recursion
        #     # sometimes??
        #     checked.append(ref)
        #     if not is_static(ref.eq, checked):
        #         return False
        # else:
        #     if not ref.static:  # TODO: and ref.static_computed? (no, that's something else)
        #         return False

        # if not ref.static or (ref.eq is not None and ref not in checked and not is_static(ref.eq, checked+ref)):
        #     return False
        #     # already determined to be static
        #     continue
        # if ref.eq is not None and not is_static(ref.eq):
        #     return False
    if isinstance(eq, reno.components.Distribution) and eq.per_timestep:
        # print("Is distribution with per_timestep, not static")
        return False
    if isinstance(
        eq,
        (reno.components.Scalar, int, float, np.ndarray),
    ):
        # keeping this explicitly to make it obvious
        # print("Equation is a scalar, int, float, or array, is static!")
        return True
    if isinstance(
        eq,
        (
            reno.components.Stock,
            reno.components.TimeRef,
            reno.components.HistoricalValue,
        ),
    ):
        # anything dealing with a stock (inherently updated every timestep) or
        # is time itself, is obv not static
        # print("Explicit time-based thing included, not static")
        return False
    if isinstance(eq, reno.ops.slice):
        # slices can potentially index time-based things (if stop is None or explicitly a
        # non-static variable) which would make this non-static
        if not eq.sub_equation_parts[0].is_static():
            return False
        if (
            eq.stop is None
            or (eq.stop is not None and not eq.stop.is_static())
            or (eq.start is not None and not eq.start.is_static())
        ):
            # print("Slice with non-static start/stop, not static")
            return False

    if isinstance(eq, reno.ops.orient_timeseries):
        return False
    if len(eq.find_refs_of_type(reno.ops.orient_timeseries)) > 0:
        return False
    # if hasattr(eq, "eq" and isinstance(eq, reno.ops.orient_timeseries):
    #     return False

    # UGH. I'm just playing whackamole (wow that word looks like guacamole) with
    # weird edgecases I keep accidentally adding.
    if (
        hasattr(eq, "eq")
        and isinstance(eq.eq, reno.components.Distribution)
        and eq.eq.per_timestep
    ):
        # print("Sub equation is a per_timestep distribution, not static")
        return False
    # this was really only created for one specific use-case, a technically
    # non-static distribution (with per_timestep specified) doesn't trigger any
    # of the above when you're calling is_static on _the containing reference_.
    # if isinstance(eq, (reno.components.Variable, reno.components.Flow)):
    #     return is_static(eq.eq, checked)
    # print("Nothing else triggered, is static")
    return True


def latex_name(name: str, cmd: str = "text") -> str:
    """Wrap a reference name in a latex string, correctly escaping anything as necessary.
    It will _not_ escape if there's a '$', indicating 'yes use mathmode' like in latex.
    This allows reference labels like `"$x_3$"`

    Example:
        >>> latex_name("testing")
        "\\text{testing}"

        >>> latex_name("x_3", "texttt")
        "\\texttt{x\\_3}"
    """
    if "$" in name:
        return f"\\{cmd}{{{name}}}"
    escaped_name = name.replace("_", "\\_")
    return f"\\{cmd}{{{escaped_name}}}"


def latex_eqline_wrap_doc(doc_text: str, highlight: bool = False) -> str:
    """Wrap a docstring in an appropriate latex block for an equation set display.

    (This handles adding a highlight color if it's a selected reference.)
    """
    string = "& \\mkern36mu "
    if highlight:
        string += "{\\color{" + DOC_HIGHLIGHT_COLOR + "} \\text{" + doc_text + "}}"
    else:
        string += "{\\color{grey} \\text{" + doc_text + "}}"
    string += " \\\\\n"
    return string


def latex_eqline_wrap(eq_text: str, highlight: bool = False) -> str:
    """Wrap an equation in an appropriate latex block for an equation set display.

    (This handles adding a highlight color if it's a selected reference.)
    """
    string = "& "
    if highlight:
        string += "{\\color{" + EQ_HIGHLIGHT_COLOR + "} " + eq_text + "}"
    else:
        string += eq_text
    string += " \\\\\n"
    return string


def range_eq_latex(min_eq, max_eq, **kwargs) -> str:
    """Get the latex string for displaying the combined min/max range of a reference."""
    if min_eq is None and max_eq is None:
        return ""

    range_eq = "\\mkern18mu \\in"

    if min_eq is None:
        min_text = "(-\\infty,"
    else:
        min_text = "[" + min_eq.latex(**kwargs) + ","

    if max_eq is None:
        max_text = "\\infty)"
    else:
        max_text = max_eq.latex(**kwargs) + "]"

    return range_eq + min_text + max_text


def get_dependency_relevant_equation_part(
    init_eqs: bool, ref: "reno.components.Reference"
) -> "reno.components.EquationPart":
    """Determine what equation part should be used to run seek_refs on
    in order to determine compute dependencies."""
    if init_eqs and isinstance(ref, reno.components.TrackedReference):
        if ref.init is None:
            # if inits were requested but no explicit init equation found
            if isinstance(ref, reno.components.Stock):
                # The init for a stock with no init just defaults to 0.0
                return reno.components.Scalar(0.0)
            # for vars and flows just use the regular equation if no init
            # specified
            return ref._implied_eq()
        return ref._implied_eq(ref.init)
    elif init_eqs and not isinstance(ref, reno.components.TrackedReference):
        # metrics would reach this I think
        # NOTE: technically this is unnecessary/handled by else below, leaving
        # it to make it explicit which condition we expect metrics to fall under
        return ref
    else:
        # NOTE: because of seek_refs behavior, for tracked references,
        # min/max ("implied eq") already gets taken care of, so returning
        # ref works in all cases.
        return ref


def dependency_compute_order(
    refs: list["reno.components.Reference"], init_eqs: bool = False, debug: bool = False
) -> list["reno.components.Reference"]:
    """Find a dependency-safe ordering for reference equations by iterating through
    and each time adding the first reference that doesn't depend on any references
    not yet added.

    This function will detect circular references (equations that depend on eachother)
    and throw an error. Primary use for this function is to correctly set order of
    equations in the pymc model/step function.

    Args:
        refs (list[reno.components.Reference]): The references to find a safe compute
            order for.
        init_eqs (bool): Use init equations to determine ordering, where available.
        debug (bool): Spit out everything it's analyzing.

    Returns:
        The dependency-ordered list of reno references passed in.
    """

    compute_order = []
    remaining = []
    iterations = 0

    # start with every ref in remaining
    remaining.extend(refs)

    # I think technically iterations should never exceed the total number of
    # references (every iteration should find one new thing that can be
    # computed), but doubling just in case
    while len(remaining) > 0 and iterations < len(refs) * 2:
        if debug:
            print(compute_order, remaining)
        ref_to_remove = None
        for obj in remaining:
            if debug:
                print("\t", obj.qual_name())
            missing_dependency = False

            eq_part = get_dependency_relevant_equation_part(init_eqs, obj)

            all_dependencies = eq_part.seek_refs()
            # disregard anything we haven't been explicitly asked about, and
            # don't include self (should only occur for seek refs of a stock
            # implied eq)
            potential_dependencies = [
                d for d in all_dependencies if d in refs and d != obj
            ]

            if debug:
                print("\t\tAll dependencies:", all_dependencies)
                print("\t\tRelevant dependencies:", potential_dependencies)
            for ref in potential_dependencies:
                if isinstance(ref, reno.components.TimeRef):
                    # time refs don't count because they have no additional
                    # dependencies and are always provided at compute time
                    continue
                if isinstance(ref, reno.components.HistoricalValue):
                    # the assumption for now is that computing a historical
                    # value's time index is static (minus timeref) and since the
                    # result was already computed in the past - no resulting
                    # refs are considered a dependency
                    continue

                if debug:
                    print("\t\t", ref)
                if ref not in compute_order:
                    if debug:
                        try:
                            print("\t\t\tCan't compute yet because of", ref.qual_name())
                        except:  # noqa: E722
                            print("???")
                    # we can't compute this obj yet, a necessary ref won't have
                    # been computed beforehand
                    missing_dependency = True
                    break

            if not missing_dependency:
                # if we made it to this point, no uncomputed refs, so we're good
                # to compute this one next
                if debug:
                    print("\t\tWe're good!")
                compute_order.append(obj)
                ref_to_remove = obj
                break
        if ref_to_remove is not None:
            remaining.remove(ref_to_remove)
        iterations += 1

    if len(remaining) > 0:
        raise Exception(
            f"Got stuck in dependency loop, could not determine compute order for {remaining}"
        )

    return compute_order


def resource_path(filename: str) -> str:
    """Get the path to the package "data resource" with matching filename. This is
    intended to get supporting js/vue files etc.

    Args:
        filename (str): the name of the resource file to get the full path of.

    Returns:
        The full package resource file path for specified filename.
    """
    path = None
    with importlib.resources.as_file(
        importlib.resources.files("reno") / "res" / filename
    ) as full_file_path:
        path = str(full_file_path)
    return path



"""Functions for generating stock & flow diagrams.

Diagrams are graphviz dot diagrams, and notably don't follow the
exact format typically used with SFDs, namely sources and sinks
aren't explicitly represented.
"""

import os

import matplotlib.pyplot as plt
import xarray as xr
from graphviz import Digraph

import reno
from reno.components import (
    Flow,
    Stock,
    TimeRef,
    TrackedReference,
    Variable,
)

# Theming constants, set for lightmode by default.
DARK_MODE = False
"""When ``True``, set graph node_attr to use colors more suited for dark backgrounds."""
SUBGRAPH_COLORS = ["#BBDDFF", "#DDBBFF"]
"""Container background colors for nested models/submodels, each subsquent
level will go one color farther into this array.

A decent option for dark mode: ``["#334455", "#443355"]``
"""
NODE_COLOR = "lightgreen"
"""Default background color to use for variable nodes.

A decent option for dark mode: ``"darkgreen"``
"""
STOCK_COLOR = "transparent"
"""Default background color to use for stock nodes."""
FLOW_COLOR = "transparent"
"""Default background color to use for flow nodes."""
EDGE_COLOR = "black"

# TODO: a set of light/dark colors with names that can be referenced
# in model.group_colors


def set_dark_mode(dark: bool = False):
    """Set themeing constants for the diagrams appropriately for
    dark/light theme.

    Note that this will also change matplotlib style so that spark plots
    match. Set the ``DARK_MODE``, ``SUBGRAPH_COLORS``, and ``NODE_COLOR``
    constants manually if this is undesired.

    Args:
        dark (bool): Pass ``False`` for light theme, ``True`` for dark theme.
    """

    global DARK_MODE, SUBGRAPH_COLORS, NODE_COLOR, EDGE_COLOR
    DARK_MODE = dark
    if not dark:
        plt.style.use("default")
        SUBGRAPH_COLORS = ["#BBDDFF", "#DDBBFF"]
        NODE_COLOR = "lightgreen"
        EDGE_COLOR = "black"
    else:
        plt.style.use("dark_background")
        SUBGRAPH_COLORS = ["#334455", "#443355"]
        NODE_COLOR = "darkgreen"
        EDGE_COLOR = "white"


def stock_flow_diagram(
    model,
    show_vars: bool = True,
    exclude_var_names: list[str] = None,
    sparklines: bool = False,
    sparkdensities: bool = False,
    sparkall: bool = False,
    g: Digraph = None,
    traces: list[xr.Dataset] = None,
    universe: list[TrackedReference] = None,
    lr: bool = False,
    hide_groups: list[str] = None,
    show_groups: list[str] = None,
    group_colors: dict[str | tuple["reno.components.TrackedReference"], str] = None,
    _level: int = 0,
) -> tuple[Digraph, list[tuple[str, str]]]:
    """Generate a graphviz dot graph for all the stocks and flows of the passed model,
    optionally including sparklines if a simulation has been run.

    Args:
        model (reno.model.Model): The model to collect stocks/flows/variables from.
        show_vars (bool): Whether to render variables in the diagram, or just stocks and flows
            (for very complex models, hiding variables can make it a bit easier to visually
            parse.)
        exclude_var_names (list[str]): Specific variables to hide in the diagram, can be used
            with ``show_vars=True`` to just show specific variables of interest.
        sparklines (bool): Draw mini graphs to the right of each stock showing plotting their
            values through time. This assumes the model has either been run, or traces are passed
            in manually with the ``traces`` argument.
        sparkdensities (bool): Draw mini density plots/histograms next to any variables that
            sample from distributions. This assumes the model has either been run, or traces
            are passed in manually with the ``traces`` argument.
        g (Digraph): Graphviz Digraph instance to render the nodes/edges on. Mostly only for
            internal use for drawing subgraphs for submodels.
        traces (list[xr.Dataset]): A list of traces or model run datasets to use for drawing
            spark plots. Each dataset will be rendered in a different color.
        universe (list[TrackedReference]): Limit rendered nodes to only those listed here, this
            includes all of stocks/flows/variables. (This acts as an initial filter, ``show_vars``
            and ``exclude_var_names`` still applies after this.)
        lr (bool): By default the graphviz plot tries to orient top-down. Specify ``True`` to
            try to orient it left-right.
        hide_groups (list[str]): A list of group/cgroup names to hide during diagramming, overriding
            model.default_hide_groups.
        show_groups (list[str]): A list of group/cgroup names to show during diagramming, overriding
            model.default_hide_groups.
        group_colors dict[str | tuple["reno.components.TrackedReference"], str]: Dictionary specifying
            colors to render groups with. An ad-hoc group defined by a tuple of references can also be
            used as a key if the appropriate cgroup does not already exist on the references.

    Returns:
        The populated Digraph instance (Jupyter can natively render this in a cell output.) and
        a list of tuples of reference connections that weren't added (usually because one of the
        references was "out of scope" part of a parent model not rendered in this one. This is
        primarily for internal use for correctly recursively plotting submodels in subgraphs.
    """
    if g is None:
        rankdir = "TB" if not lr else "LR"
        if DARK_MODE:
            g = Digraph(
                name=model.name,
                graph_attr=dict(rankdir=rankdir, bgcolor="#181818"),
                node_attr=dict(
                    fontcolor="#e6e6e6",
                    style="filled",
                    color="#e6e6e6",
                    fillcolor="#333333",
                ),
                edge_attr=dict(color="#e6e6e6", fontcolor="#e6e6e6"),
            )
        else:
            g = Digraph(name=model.name, graph_attr=dict(rankdir=rankdir))
    if exclude_var_names is None:
        exclude_var_names = []

    g, out_of_scope_stock_connections = add_stocks(
        g,
        [stock for stock in model.stocks if not stock.implicit],
        sparklines=sparklines,
        traces=traces,
        universe=universe,
        show_vars=show_vars,
        hide_groups=hide_groups,
        show_groups=show_groups,
        group_colors=group_colors,
    )
    out_of_scope_var_connections = []
    if show_vars:
        variables = filter_variables(
            [var for var in model.vars if not var.implicit], exclude_var_names
        )
        g, out_of_scope_var_connections = add_vars(
            g,
            variables,
            exclude_var_names,
            sparkdensities,
            sparkall,
            traces,
            universe=universe,
            hide_groups=hide_groups,
            show_groups=show_groups,
            group_colors=group_colors,
        )
    else:
        variables = []
    g, out_of_scope_flow_connections = add_flows(
        g,
        [flow for flow in model.flows if not flow.implicit],
        variables,
        sparkall,
        traces,
        universe=universe,
        hide_groups=hide_groups,
        show_groups=show_groups,
        group_colors=group_colors,
    )

    missing_connections = (
        out_of_scope_var_connections
        + out_of_scope_flow_connections
        + out_of_scope_stock_connections
    )

    # recursively render any submodels.
    for submodel in model.models:
        with g.subgraph(name=submodel.name) as g_s:
            g_s.attr(style="filled")
            g_s.attr(color=SUBGRAPH_COLORS[_level])
            g_s.attr(cluster="true")
            g_s.attr(label=submodel.label)
            g_s.attr(fontcolor="#888888")
            m = getattr(model, submodel.name)
            unused_graph, remaining_connections = stock_flow_diagram(
                m,
                show_vars=show_vars,
                exclude_var_names=exclude_var_names,
                sparklines=sparklines,
                sparkdensities=sparkdensities,
                sparkall=sparkall,
                g=g_s,
                traces=traces,
                universe=universe,
                lr=lr,
                hide_groups=hide_groups,
                show_groups=show_groups,
                group_colors=group_colors,
                _level=_level + 1,
            )
            for src, dst in remaining_connections:
                # check if we are at the correct level model to draw this,
                # or do we pass it on up?
                if src.model == model or dst.model == model:
                    if universe is not None and (
                        src not in universe or dst not in universe
                    ):
                        continue
                    if src in exclude_var_names or dst in exclude_var_names:
                        continue
                    draw_appropriate_edge(g, src, dst)
                else:
                    missing_connections.append((src, dst))

    return g, missing_connections


def add_stock_io_edge(
    g: Digraph, stock: Stock, flow: Flow, dir: str = None, group_colors: dict = None
):
    """Edge with attributes for heavily highlighting in flows and
    outflows from stocks."""

    if flow.implicit and flow in stock.in_flows:
        for ref in flow.seek_refs():
            if isinstance(ref, Flow):
                add_stock_io_edge(g, stock, ref, "in", group_colors)
        return
    elif flow.implicit and flow in stock.out_flows:
        for ref in flow.seek_refs():
            if isinstance(ref, Flow):
                add_stock_io_edge(g, stock, ref, "out", group_colors)
        return

    color = EDGE_COLOR
    if (
        get_reference_color(flow, no_default=True, group_colors=group_colors)
        is not None
    ):
        color = get_reference_color(flow, group_colors=group_colors)
    elif (
        get_reference_color(stock, no_default=True, group_colors=group_colors)
        is not None
    ):
        color = get_reference_color(stock, group_colors=group_colors)
    if flow in stock.in_flows or dir == "in":
        name1 = flow.qual_name()
        name2 = stock.qual_name()
    elif flow in stock.out_flows or dir == "out":
        name1 = stock.qual_name()
        name2 = flow.qual_name()
    else:
        raise Exception(
            f"Could not draw edge for flow {flow.qual_name()} and stock {stock.qual_name()}, stock does not use flow."
        )
    g.edge(name1, name2, style="bold", weight="50", color=color)


def add_stock_io_like_edge(
    g: Digraph, to_flow: Flow, in_flow: Flow, group_colors: dict = None
):
    """Allow a stock-io-like edge between two flows (if one is explicitly listed
    as an inflow to another flow)"""
    color = EDGE_COLOR
    if (
        get_reference_color(in_flow, no_default=True, group_colors=group_colors)
        is not None
    ):
        color = get_reference_color(in_flow, group_colors=group_colors)
    elif (
        get_reference_color(to_flow, no_default=True, group_colors=group_colors)
        is not None
    ):
        color = get_reference_color(to_flow, group_colors=group_colors)
    name1 = in_flow.qual_name()
    name2 = to_flow.qual_name()
    g.edge(name1, name2, style="bold", weight="50", color=color)


def add_stock_limit_edge(g: Digraph, ref, stock: Stock):
    """De-emphasize stock variable/flow connections if they're only
    used in min/max constraints."""
    g.edge(
        ref.qual_name(),
        stock.qual_name(),
        style="dotted",
        arrowsize=".5",
    )


def add_to_flow_edge(g: Digraph, ref, flow: Flow, deemphasize: bool = False):
    """Non-stock-inflow/outflow-related edges that point to flows
    are slightly de-emphasized."""
    style = "dotted" if isinstance(ref, Variable) else "dashed"
    constraint = "false" if isinstance(ref, Stock) or deemphasize else "true"
    weight = "1" if not deemphasize else "0"
    # constraint = "false"
    # TODO: constraint should maybe take same model scope into account?
    # (stock to out flows edges already handled, any others
    # shouldn't affect rank)
    g.edge(
        ref.qual_name(),
        flow.qual_name(),
        style=style,
        arrowsize=".5",
        constraint=constraint,
        weight=weight,
    )


def add_to_var_edge(g: Digraph, ref, variable: Variable):
    """Edges pointing at variables are de-emphasized (usually these are
    auxiliary variables - variables calculated based on other variables.)"""
    g.edge(
        ref.qual_name(),
        variable.qual_name(),
        style="dotted",
        arrowsize=".5",
    )


# this is mostly only important for determining how to draw edges between refs
# in different scopes
# TODO: pretty sure there's still a mistake in this, will sometimes still try to
# include out of scope variable and just gets added as default ellipse.
def draw_appropriate_edge(
    g: Digraph, src: TrackedReference, dst: TrackedReference, group_colors: dict = None
):
    """Switch between edge types correctly based on the types of src/dst nodes."""
    if (
        isinstance(src, Flow)
        and isinstance(dst, Stock)
        and (src in dst.in_flows or src in dst.out_flows)
    ):
        add_stock_io_edge(g, dst, src, group_colors=group_colors)
    elif (
        isinstance(src, Stock)
        and isinstance(dst, Flow)
        and (dst in src.in_flows or dst in src.out_flows)
    ):
        add_stock_io_edge(g, src, dst, group_colors=group_colors)
    elif isinstance(dst, Stock) and (src in dst.min_refs() or src in dst.max_refs()):
        print("INCORPORAINT", dst, src)
        add_stock_limit_edge(g, src, dst)
    elif isinstance(dst, Flow):
        add_to_flow_edge(g, src, dst)
    elif isinstance(dst, Variable):
        add_to_var_edge(g, src, dst)
    else:
        raise Exception(
            f"What type of nodes are you asking me to connect??? {src} -> {dst}?"
        )


def filter_variables(
    all_variables: list[Variable], exclude_names: list[str]
) -> list[Variable]:
    """Take a list of variables and a list of names of variables to remove from the list.

    Returns the list of variables that do NOT have a name in the exclude_names list.
    """
    filtered = [
        variable for variable in all_variables if variable.name not in exclude_names
    ]
    return filtered


def get_reference_color(  # noqa: C901
    ref, no_default: bool = False, group_colors: dict = None
) -> str:
    # TODO: dark/light handling
    if isinstance(ref, reno.Stock):
        color = STOCK_COLOR
    elif isinstance(ref, reno.Variable):
        color = NODE_COLOR
    elif isinstance(ref, reno.Flow):
        color = FLOW_COLOR
    if no_default:
        color = None

    if group_colors is None:
        group_colors = {}

    # manually specified group_colors take precedence
    for group in group_colors:
        if isinstance(group, tuple):
            if ref in group:
                return group_colors[group]

    # cgroup takes next precedence
    if isinstance(ref.cgroup, list):
        for cgroup in ref.cgroup:
            # manually specified takes precedence
            if cgroup in group_colors:
                return group_colors[cgroup]

            # otherwise check model default
            if cgroup in ref.model.group_colors:
                return ref.model.group_colors[cgroup]

    # model-defined defaults take precedence next, with cgroup over group
    if ref.cgroup != "":
        if ref.cgroup in group_colors:
            return group_colors[ref.cgroup]
        if ref.cgroup in ref.model.group_colors:
            return ref.model.group_colors[ref.cgroup]
    if ref.group != "":
        if ref.group in group_colors:
            return group_colors[ref.group]
        if ref.group in ref.model.group_colors:
            return ref.model.group_colors[ref.group]
    return color


def should_render(
    ref, universe, hide_groups: list[str] = None, show_groups: list[str] = None
) -> bool:
    if show_groups is None:
        show_groups = []
    if hide_groups is None:
        hide_groups = []

    if universe is not None and ref not in universe:
        return False
    # if isinstance(ref, (TimeRef, HistoricalValue)):
    if not isinstance(ref, (Stock, Flow, Variable)):
        return False
    # show_groups takes precedence over hide_groups
    if isinstance(ref.cgroup, list):
        # manually passed takes precedence
        found_hide = False
        for cgroup in ref.cgroup:
            if cgroup in show_groups:
                return True
            if cgroup in hide_groups:
                found_hide = True
        if found_hide:
            return False

        # otherwise look at default
        for cgroup in ref.model.default_hide_groups:
            if cgroup in ref.cgroup:
                return False
    if ref.cgroup in show_groups or ref.group in show_groups:
        return True
    if ref.cgroup in hide_groups or ref.group in hide_groups:
        return False
    if (
        ref.cgroup in ref.model.default_hide_groups
        or ref.group in ref.model.default_hide_groups
    ):
        return False
    # somehow breaks add_stock_io_edge when this is on??
    # if hasattr(ref, "implicit") and ref.implicit:
    #     return False
    return True


def add_stocks(  # noqa: C901
    g: Digraph,
    stocks: list[Stock],
    sparklines: bool = False,
    traces: list[xr.Dataset] = None,
    universe: list[TrackedReference] = None,
    show_vars: bool = True,
    hide_groups: list[str] = None,
    show_groups: list[str] = None,
    group_colors: dict = None,
):
    """Add nodes and edges for all passed stocks to the passed graph.

    Args:
        sparklines (bool): include small little line graph of each stock in the
            diagram. This only works if a simulation has been run.
    """

    render_in_parent_scope = []

    for stock in stocks:
        if not should_render(stock, universe, hide_groups, show_groups):
            continue

        # add a node for each stock
        if not sparklines:
            g.node(
                name=stock.qual_name(),
                label=stock.label,
                shape="rect",
                group=stock.group,
                style="filled",
                fillcolor=get_reference_color(stock, group_colors=group_colors),
            )
        else:
            # to force the sparkline graph image node to render right next to
            # its stock node, put them both in a subgraph with rank=same
            with g.subgraph(graph_attr={"rank": "same", "cluster": "false"}) as c:
                c.node(
                    name=stock.qual_name(),
                    label=stock.label,
                    shape="rect",
                    group=stock.group,
                    style="filled",
                    fillcolor=get_reference_color(stock, group_colors=group_colors),
                )

                # generate the sparkline graph
                with plt.ioff():
                    fig, ax = plt.subplots(figsize=(2, 1))
                    # use the bayes one if available
                    if stock.model.trace is not None or traces is not None:
                        if traces is None:
                            traces = [stock.model.trace.prior]
                            if "posterior" in stock.model.trace:
                                traces.append(stock.model.trace.posterior)

                        reno.viz.compare_seq(
                            stock.qual_name(), traces, ax=ax, legend=False, title=""
                        )
                    else:
                        if traces is None:
                            ds = stock.model.dataset()
                            traces = [ds]
                        reno.viz.compare_seq(
                            stock.qual_name(), traces, ax=ax, legend=False, title=""
                        )
                    ax.xaxis.set_ticks([])
                    fig.tight_layout()
                    os.makedirs(".plotcache", exist_ok=True)
                    fig.savefig(f".plotcache/{stock.qual_name()}.png")
                    plt.close(fig)

                c.node(
                    name=f"{stock.qual_name()}_fig",
                    label="",
                    image=f".plotcache/{stock.qual_name()}.png",
                    shape="none",
                    group=stock.group,
                )
                c.edge(
                    stock.qual_name(),
                    f"{stock.qual_name()}_fig",
                    constraint="true",
                    # weight="20",
                    # weight="200",
                    dir="none",
                )

        # look at each stock's inflows/outflows and add the appropriate edges
        # (this works even if the flow "nodes" haven't been added yet.)
        handled_refs = []
        for flow in stock.in_flows:
            if not should_render(flow, universe, hide_groups, show_groups):
                continue
            if reno.utils.is_ref_in_parent_scope(flow, stock):
                # add refs themselves not qualname, need to access model
                render_in_parent_scope.append((flow, stock))
                continue
            add_stock_io_edge(g, stock, flow, group_colors=group_colors)
            handled_refs.append(flow.qual_name())
        for flow in stock.out_flows:
            if not should_render(flow, universe, hide_groups, show_groups):
                continue
            if reno.utils.is_ref_in_parent_scope(flow, stock):
                # add refs themselves not qualname, need to access model
                render_in_parent_scope.append((flow, stock))
                continue
            add_stock_io_edge(g, stock, flow, group_colors=group_colors)
            handled_refs.append(flow.qual_name())
        remaining = []
        remaining.extend(stock.min_refs())
        remaining.extend(stock.max_refs())
        for ref in remaining:
            if isinstance(ref, Variable) and not show_vars:
                continue
            if ref not in handled_refs:
                if not should_render(ref, universe, hide_groups, show_groups):
                    continue
                if reno.utils.is_ref_in_parent_scope(ref, stock):
                    # add refs themselves not qualname, need to access model
                    render_in_parent_scope.append((ref, stock))
                    continue
                add_stock_limit_edge(g, ref, stock)
    return g, render_in_parent_scope


def add_vars(
    g: Digraph,
    variables: list[Variable],
    exclude_vars: list[str],
    sparkdensities: bool = False,
    sparkall: bool = False,
    traces: list[xr.Dataset] = None,
    universe: list[TrackedReference] = None,
    hide_groups: list[str] = None,
    show_groups: list[str] = None,
    group_colors: dict = None,
):
    """Add variables and edges between variables to the passed graphviz graph."""
    rendered_edges = []
    render_in_parent_scope = []
    for variable in variables:
        # if universe is not None and variable not in universe:
        #     continue
        if not should_render(variable, universe, hide_groups, show_groups):
            continue
        if not sparkdensities or (
            sparkdensities
            and (variable.qual_name() not in variable.model.trace_RVs and not sparkall)
        ):
            g.node(
                name=variable.qual_name(),
                label=variable.label,
                style="rounded,filled",
                fillcolor=get_reference_color(variable, group_colors=group_colors),
                shape="rect",
                fontsize="10pt",
                height=".2",
                group=variable.group,
            )
        else:
            # to force the sparkdensity graph image node to render right next to
            # its variable node, put them both in a subgraph with rank=same
            with g.subgraph(graph_attr={"rank": "same"}) as c:
                c.node(
                    name=variable.qual_name(),
                    label=variable.label,
                    style="rounded,filled",
                    fillcolor=get_reference_color(variable, group_colors=group_colors),
                    shape="rect",
                    fontsize="10pt",
                    height=".2",
                    group=variable.group,
                )

                # generate sparkdensity graph
                with plt.ioff():
                    fig, ax = plt.subplots(figsize=(2, 1))

                    if traces is None:
                        traces = [variable.model.trace.prior]
                        if "posterior" in variable.model.trace:
                            traces.append(variable.model.trace.posterior)

                    reno.viz.compare_posterior(
                        variable.qual_name(), traces, ax=ax, legend=False, title=""
                    )

                    ax.yaxis.set_ticks([])
                    fig.tight_layout()
                    os.makedirs(".plotcache", exist_ok=True)
                    fig.savefig(f".plotcache/{variable.qual_name()}.png")
                    plt.close(fig)

                c.node(
                    name=f"{variable.qual_name()}_fig",
                    label="",
                    image=f".plotcache/{variable.qual_name()}.png",
                    shape="none",
                    group=variable.group,
                )
                c.edge(
                    variable.qual_name(),
                    f"{variable.qual_name()}_fig",
                    constraint="true",
                    weight="20",
                    dir="none",
                )

        # add any edges between variables
        # -- unclear if correct --
        for ref in variable.seek_refs():
            if not should_render(ref, universe, hide_groups, show_groups):
                continue
            if hasattr(ref, "implicit") and ref.implicit:
                continue
            if (
                not isinstance(ref, reno.components.TimeRef)
                and ref.name not in exclude_vars
                and (ref.qual_name(), variable.qual_name()) not in rendered_edges
            ):
                # don't render connections to things in parent model - parent
                # model should handle
                if reno.utils.is_ref_in_parent_scope(ref, variable):
                    # add refs themselves not qualname, need to access model
                    render_in_parent_scope.append((ref, variable))
                    continue

                add_to_var_edge(g, ref, variable)
                rendered_edges.append((ref.qual_name(), variable.qual_name()))
        # -- /unclear if correct --
    return g, render_in_parent_scope


def add_flows(
    g: Digraph,
    flows: list[Flow],
    variables: list[Variable],
    sparkall: bool = False,
    traces: list[xr.Dataset] = None,
    universe: list[TrackedReference] = None,
    hide_groups: list[str] = None,
    show_groups: list[str] = None,
    group_colors: dict[str | list["reno.components.TrackedReference"], str] = None,
):
    """Add flows and edges from variables to the passed graphviz graph."""
    rendered_edges = []

    render_in_parent_scope = []

    for flow in flows:
        if not should_render(flow, universe, hide_groups, show_groups):
            continue
        if not sparkall:
            g.node(
                name=flow.qual_name(),
                label=flow.name,
                shape="plain",
                group=flow.group,
                style="filled",
                fillcolor=get_reference_color(flow, group_colors=group_colors),
            )
        else:
            # to force the sparkline graph image node to render right next to
            # its flow node, put them both in a subgraph with rank=same
            with g.subgraph(graph_attr={"rank": "same"}) as c:
                c.node(
                    name=flow.qual_name(),
                    label=flow.name,
                    shape="plain",
                    group=flow.group,
                    style="filled",
                    fillcolor=get_reference_color(flow, group_colors=group_colors),
                )

                # generate the sparkline graph
                with plt.ioff():
                    fig, ax = plt.subplots(figsize=(2, 1))
                    # use the bayes one if available
                    if flow.model.trace is not None or traces is not None:
                        if traces is None:
                            traces = [flow.model.trace.prior]
                            if "posterior" in flow.model.trace:
                                traces.append(flow.model.trace.posterior)

                        reno.viz.compare_seq(
                            flow.qual_name(), traces, ax=ax, legend=False, title=""
                        )
                    else:
                        # TODO: can probably still use compare_seq for regular
                        # reno run plot? (since traces can be xrDataset)
                        # TODO: TODO: TODO: flow.plot doesn't exist?!
                        # flow.plot(ax)
                        if traces is None:
                            ds = flow.model.dataset()
                            traces = [ds]
                        reno.viz.compare_seq(
                            flow.qual_name(), traces, ax=ax, legend=False, title=""
                        )
                    ax.xaxis.set_ticks([])
                    fig.tight_layout()
                    os.makedirs(".plotcache", exist_ok=True)
                    fig.savefig(f".plotcache/{flow.qual_name()}.png")
                    plt.close(fig)

                c.node(
                    name=f"{flow.qual_name()}_fig",
                    label="",
                    image=f".plotcache/{flow.qual_name()}.png",
                    shape="none",
                    group=flow.group,
                )
                c.edge(
                    flow.qual_name(),
                    f"{flow.qual_name()}_fig",
                    constraint="true",
                    weight="20",
                    dir="none",
                )

        for ref, ref_types in flow.seek_refs(include_ref_types=True).items():
            if not should_render(ref, universe, hide_groups, show_groups):
                continue
            if hasattr(ref, "implicit") and ref.implicit:
                continue
            if (
                (
                    ref in variables
                    # TODO: a function called before this that removes variables
                    # of name x
                    and (ref.qual_name(), flow.qual_name()) not in rendered_edges
                )
                # TODO: ...what?
                or (
                    ref not in variables
                    and not isinstance(
                        ref, (TimeRef, Variable)
                    )  # TODO: Variable wasn't here before, with this here means we won't be able to show connections between submodels?
                    and (ref.qual_name(), flow.qual_name()) not in rendered_edges
                    # don't re-render a line from a stock to a flow if the flow
                    # is explicitly an outflow of the stock
                    and not (isinstance(ref, Stock) and flow in ref.out_flows)
                )
            ):
                # don't render connections to things in parent model - parent
                # model should handle
                if reno.utils.is_ref_in_parent_scope(ref, flow):
                    # add refs themselves not qualname, need to access model
                    render_in_parent_scope.append((ref, flow))
                    continue

                if "inflow" in ref_types:
                    add_stock_io_like_edge(g, flow, ref, group_colors=group_colors)
                else:
                    add_to_flow_edge(g, ref, flow, deemphasize="outflows" in ref_types)
                rendered_edges.append((ref.qual_name(), flow.qual_name()))

    return g, render_in_parent_scope



"""Various visualization tools, classes, and functions.

A lot of this may eventually need to be broken out into separate
submodules for sanity."""

import math
from collections.abc import Callable
from dataclasses import dataclass

import arviz as az
import ipywidgets as ipw
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import panel as pn
import xarray as xr
from IPython.display import clear_output, display
from ipywidgets import Layout
from matplotlib import ticker
from matplotlib.collections import LineCollection
from matplotlib.lines import Line2D
from matplotlib.transforms import ScaledTranslation
from scipy.stats import gaussian_kde

import reno
from reno.components import Flow, Scalar, Stock, Variable
from reno.interactive_latex import InteractiveLatex
from reno.parser import parse
from reno.utils import latex_eqline_wrap, latex_eqline_wrap_doc


def _create_seq_line_collection(seq_np_values: np.ndarray, **kwargs) -> LineCollection:
    """Running ax.plot for each individual line (of potentially thousands) is very slow,
    but you can create a LineCollection of all of the lines and render almost instantly,
    see https://matplotlib.org/stable/gallery/shapes_and_collections/line_collection.html

    Args:
        seq_np_values (np.ndarray): A matrix of y-values to plot (assumes that each row
            is a timeseries.)
        **kwargs: Any additional arguments passed to matplotlib.collections.LineCollection

    Returns:
        A LineCollection object that can be plotted on an axis with ``ax.add_collection()``
    """

    # the trick is that LineCollection needs to be a list of lines where each
    # line is a list of (x0, y0), ... coords, e.g.
    # [[[x0, y0], [x1, y1]...], ...,]

    # so, we do some numpy magic to create the x-domain for these sequences and
    # shape the matrix of y data appropriately

    domain_size = seq_np_values.shape[-1]
    # make sure to flatten if it wasn't passed flattened already
    seqs = seq_np_values.reshape([-1, domain_size])
    # create a corresponding domain (0, 1, 2, ...) for each sequence
    domains = np.tile(np.arange(0, domain_size, 1), (seqs.shape[0], 1))
    # combine!
    line_data = np.stack((domains, seqs), axis=2)
    # at this point we should have (num_lines, timesteps, 2) where first of 2 is domain)

    return LineCollection(line_data, **kwargs)


def _get_label_and_dataset(
    trace_collection: (
        list[az.InferenceData | xr.Dataset] | dict[str, az.InferenceData | xr.Dataset]
    ),
    key_or_trace: str | az.InferenceData | xr.Dataset,
) -> tuple[str, xr.Dataset]:
    """Traces can be passed to the comparison functions multiple ways, so this is
    a method to consistently get a label for a trace and the desired dataset."""
    # determine if label explicitly provided via dictionary or if we need
    # to auto-assign
    if isinstance(trace_collection, dict):
        label = key_or_trace
        trace = trace_collection[key_or_trace]
    else:
        label = f"trace {trace_collection.index(key_or_trace)}"
        trace = key_or_trace

    # automatically grab the posterior if not explicitly requested
    if isinstance(trace, az.InferenceData):
        ds = trace.posterior
    else:
        ds = trace

    return label, ds


def _get_sample_count(array: xr.DataArray) -> int:
    """Dimensions for base reno and pymc are different since pymc has chains, so this
    throws off alpha calculations, use this to get the number of things that will be drawn.
    """
    if "chain" in array.coords:
        array = array.stack(sample=["chain", "draw"])
    return len(array.coords["sample"])


def compare_seq(
    varname: str,
    traces: (
        list[az.InferenceData | xr.Dataset] | dict[str, az.InferenceData | xr.Dataset]
    ),
    prior_trace=None,
    ax=None,
    legend: bool = True,
    title: str = None,
    **figargs,
):
    """Plot the timeseries data for the specified variable from each of the passed traces + prior trace.

    Args:
        varname (str): The name of the variable in the xr.Datasets to plot
        traces (list[az.InferenceData | xr.Dataset] | dict[str, az.InferenceData | xr.Dataset): A list
            or dictionary of traces or dataset to plot the variable from. Passed traces will plot from
            the posterior, pass the specific dataset if you need the priors (``pymc_trace.prior``). If
            a dictionary is used, the legend will use the specified keys.
        prior_trace (az.InferenceData | xr.Dataset): A trace or dataset to plot with the 'prior' key.
        ax: Optionally pass an axis if one already exists, otherwise this function will create a new one,
            using any additional figargs passed.
        legend (bool): Whether to render a legend or not. If ``True`` and traces is a dictionary, the
            keys will be used as the legend labels.
        title (str): Optional title to set on the axis.
        **figargs: Parameters to pass to ``plt.subplots(**figargs)`` if no axis passed in.
    """
    if ax is None:
        with plt.ioff():
            fig, ax = plt.subplots(**figargs)

    cat_col = 0

    legend_handles = []

    if prior_trace is not None:
        alpha = 0.01 if _get_sample_count(prior_trace.prior[varname]) > 10 else 0.75
        ax.add_collection(
            _create_seq_line_collection(
                prior_trace.prior[varname].values, color=f"C{cat_col}", alpha=alpha
            )
        )
        legend_handles.append(Line2D([0], [0], label="prior", color=f"C{cat_col}"))
        cat_col += 1

    for i, trace in enumerate(traces):
        label, ds = _get_label_and_dataset(traces, trace)

        if varname not in ds:
            continue
        alpha = 0.01 if _get_sample_count(ds[varname]) > 10 else 0.75
        ax.add_collection(
            _create_seq_line_collection(
                ds[varname].values, color=f"C{cat_col}", alpha=alpha
            )
        )
        legend_handles.append(Line2D([0], [0], label=label, color=f"C{cat_col}"))
        cat_col += 1

    if legend:
        ax.legend(handles=legend_handles)

    if title is None:
        ax.set_title(varname)
    else:
        ax.set_title(title)

    ax.autoscale()
    return ax.get_figure()


def _plot_posterior_values(
    ax,
    values: np.ndarray,
    label: str,
    smoothing: float = 0.1,
    num_traces_to_plot: int = 1,
    trace_index: int = 0,
):
    """num_traces_to_plot only matters for the bar chart, for handling offset."""
    # plot bars for categorical data
    if values.dtype in ("int64", "int32", "int8"):
        width = 0.75 / num_traces_to_plot

        counts = pd.Series(values.flatten()).value_counts()
        # normalize
        counts = counts / counts.sum()
        x = counts.index.values + width * trace_index
        ax.bar(x, counts.values, width=width - 0.02, label=label, alpha=0.75)
        # ax.bar_label(rects)
        ax.set_xticks(counts.index.values + width * (num_traces_to_plot - 1) / 2)
        ax.set_xticklabels(counts.index.values)
    elif len(values.flatten()) == 1:
        ax.axvline(x=values.flatten()[0], label=label, color=f"C{trace_index}")
    else:
        domain, densities = density(values.flatten(), smoothing)
        ax.plot(domain, densities, label=label)
        ax.fill_between(domain, densities, alpha=0.1)


def compare_posterior(
    varname: str,
    traces: (
        list[az.InferenceData | xr.Dataset] | dict[str, az.InferenceData | xr.Dataset]
    ),
    prior_trace=None,
    smoothing: float = 0.1,
    ax=None,
    legend: bool = True,
    title=None,
    per_dim=None,
    **figargs,
):
    """Plot the sampled distribution densities for the specified variable from each of the passed traces + prior trace.

    Args:
        varname (str): The name of the variable in the xr.Datasets to plot
        traces (list[az.InferenceData | xr.Dataset] | dict[str, az.InferenceData | xr.Dataset): A list
            or dictionary of traces or dataset to plot the variable from. Passed traces will plot from
            the posterior, pass the specific dataset if you need the priors (``pymc_trace.prior``). If
            a dictionary is used, the legend will use the specified keys.
        prior_trace (az.InferenceData | xr.Dataset): A trace or dataset to plot with the 'prior' key.
        smoothing (float): What degree of smoothing to apply to the density plot. Lower = more bumpy.
        ax: Optionally pass an axis if one already exists, otherwise this function will create a new one,
            using any additional figargs passed.
        legend (bool): Whether to render a legend or not. If ``True`` and traces is a dictionary, the
            keys will be used as the legend labels.
        title (str): Optional title to set on the axis.
        per_dim (str): TODO
        **figargs: Parameters to pass to ``plt.subplots(**figargs)`` if no axis passed in.
    """
    # NOTE: use per_dim to either individually plot chains, or individually plot
    # dirichlet components, None to entirely flatten (the default)
    # TODO: (not implemented yet)

    if ax is None:
        with plt.ioff():
            fig, ax = plt.subplots(**figargs)

    num_traces = len(traces)
    trace_index = 0
    if prior_trace is not None:
        num_traces += 1

    if prior_trace is not None:
        _plot_posterior_values(
            ax,
            prior_trace.prior[varname].values,
            "prior",
            smoothing,
            num_traces,
            trace_index,
        )
        trace_index += 1

    for i, trace in enumerate(traces):
        label, ds = _get_label_and_dataset(traces, trace)

        if varname not in ds:
            continue

        _plot_posterior_values(
            ax, ds[varname].values, label, smoothing, num_traces, trace_index
        )
        trace_index += 1

    if legend:
        ax.legend()

    if title is None:
        ax.set_title(varname)
    else:
        ax.set_title(title)

    return ax.get_figure()


def plot_refs_single_axis(
    trace: az.InferenceData | xr.Dataset,
    ref_list: list[str | reno.components.Reference],
    num_ticks: int = 3,
    # ax = None,
    **figargs,
):
    """I've seen this type of plot in a few SDM textbooks and references at this point,
    it plots all the specified references on the same plot and scales each one individually.

    This means there's a separate y-axis per reference, and getting the tick labels to not
    conflict with eachother and be offset was a bit of a trick, so this simplifies that logic.
    """
    plt.ioff()
    fig, ax1 = plt.subplots(**figargs)
    plt.ion()

    refs = [
        ref.qual_name() if isinstance(ref, reno.components.Reference) else ref
        for ref in ref_list
    ]

    if isinstance(trace, az.InferenceData):
        ds = trace.posterior
    else:
        ds = trace

    def label_offset(n, n_max):
        """Algorithm to vertically offset each label for a single tick so they
        don't overlap. n is the index of the current label, n_max is the total
        number of labels that have to fit."""
        spacing = 5
        total = (n_max - 1) * spacing * 2
        label_n = ((n_max - n) * spacing * 2) - (total / 2)
        return label_n

    legend_handles = []
    for i, ref in enumerate(refs):
        alpha = 0.01 if _get_sample_count(ds[ref]) > 10 else 0.75

        if i == 0:
            ax = ax1
        else:
            ax = ax1.twinx()

        ax.add_collection(
            _create_seq_line_collection(ds[ref].values, color=f"C{i}", alpha=alpha)
        )
        ax.autoscale()
        ax.yaxis.set_major_locator(ticker.LinearLocator(num_ticks))
        ax.spines["left"].set_position(("axes", 0.0))
        ax.spines["left"].set_visible(True)
        ax.yaxis.set_label_position("left")
        ax.yaxis.set_ticks_position("left")
        ax.tick_params(axis="y", colors=f"C{i}")

        for tick in ax.yaxis.get_majorticklabels():
            tick.set_transform(
                tick.get_transform()
                + ScaledTranslation(
                    0 / 72, label_offset(i + 1, len(refs)) / 72, fig.dpi_scale_trans
                )
            )

        legend_handles.append(Line2D([0], [0], label=ref, color=f"C{i}"))

    ax1.legend(handles=legend_handles)
    return fig


def plot_trace_refs(
    reference_model,
    traces: (
        list[az.InferenceData | xr.Dataset] | dict[str, az.InferenceData | xr.Dataset]
    ),
    ref_list: list[str | reno.components.Reference],
    cols: int = None,
    rows: int = None,
    smoothing: float = 0.1,
    **figargs,
):
    """Create a set of plots for each of the specified references, automatically plotting timeseries
    sequences or distribution densities where relevant.

    Args:
        reference_model: The model from which the datasets are being plotted, this is used to help
            axes titles and determine which references are random variables etc.
        traces (list[az.InferenceData | xr.Dataset] | dict[str, az.InferenceData | xr.Dataset): A list
            or dictionary of traces or dataset to plot the references from. Passed traces will plot from
            the posterior, pass the specific dataset if you need the priors (``pymc_trace.prior``). If
            a dictionary is used, the legends will use the specified keys.
        ref_list (list[str | reno.components.Reference]): Either the string names or actual reference
            objects to plot the data of. (A separate plot will be created for each reference.)
        cols (int): The number of columns to split the plots into.
        rows (int): The number of rows to split the plots into.
        smoothing (float): Degree of smoothing to apply to density plots, lower = more bumpy.
        **figargs: Arguments to pass to the figure creation ``plt.subplots(..., **figargs)`` call.
    """
    # remaining_refs = [*ref_list]
    remaining_refs = [
        ref.qual_name() if isinstance(ref, reno.components.Reference) else ref
        for ref in ref_list
    ]
    num_things = len(ref_list)

    if cols is None and num_things > 1:
        cols = 2
    elif cols is None:
        cols = 1
    if rows is None:
        rows = math.ceil(num_things / cols)
    if rows == 0:
        rows = 1

    with plt.ioff():
        fig, axs = plt.subplots(rows, cols, **figargs)

    if rows == 1 and cols == 1:
        axs = [[axs]]
    elif rows == 1:
        axs = [axs]
    elif cols == 1:
        new_axs = []
        for ax in axs:
            new_axs.append([ax])
        axs = new_axs

    row_i = 0
    col_i = 0

    def advance_indices():
        nonlocal row_i, col_i
        col_i += 1
        if col_i >= cols:
            col_i = 0
            row_i += 1

    # RVs
    rv_names = [
        name
        for name in reference_model.trace_RVs
        if not name.endswith("_likelihood") and name in remaining_refs
    ]
    rvs = [ref for ref in reference_model.all_refs() if ref.qual_name() in rv_names]
    for rv in rvs:
        compare_posterior(
            rv.qual_name(),
            traces,
            title=f"RV: {rv.qual_name()}",
            smoothing=smoothing,
            ax=axs[row_i][col_i],
            **figargs,
        )
        advance_indices()
        remaining_refs.remove(rv.qual_name())

    # other variables
    static_vars = []
    nonstatic_vars = []
    for var in reference_model.all_vars():
        if var.qual_name() in remaining_refs:
            if var.is_static():
                static_vars.append(var)
            else:
                nonstatic_vars.append(var)

    for var in static_vars:
        compare_posterior(
            var.qual_name(),
            traces,
            smoothing=smoothing,
            title=f"Var: {var.qual_name()}",
            ax=axs[row_i][col_i],
            **figargs,
        )
        advance_indices()
        remaining_refs.remove(var.qual_name())

    # stocks/flows
    seqs = [
        ref for ref in reference_model.all_refs() if ref.qual_name() in remaining_refs
    ]
    # seq_names.extend(nonstatic_var_names)  # already handled by what wouldn't
    # be removed from remaining_refs and all_refs
    title_prefix = {Stock: "Stock", Flow: "Flow", Variable: "Var"}
    for ref in seqs:
        title = f"{title_prefix[type(ref)]}: {ref.qual_name()}"
        compare_seq(
            ref.qual_name(), traces, ax=axs[row_i][col_i], title=title, **figargs
        )
        advance_indices()
        remaining_refs.remove(ref.qual_name())

    # metrics
    metrics = [
        metric
        for metric in reference_model.all_metrics()
        if metric.name in remaining_refs
    ]
    for metric in metrics:
        name = metric.qual_name()
        compare_posterior(
            name,
            traces,
            smoothing=smoothing,
            ax=axs[row_i][col_i],
            title=f"Metric: {name}",
        )
        advance_indices()
        remaining_refs.remove(name)
    fig.tight_layout()
    return fig


def density(data: np.ndarray, smoothing: float = 0.1) -> tuple[np.ndarray, np.ndarray]:
    """Get smoothed histogram (density plot) data from an array.

    Returns a tuple of the domain (x) and the smoothed values (y).
    """
    # convert data to floats and add jitter if all the same?
    # if data.var() == 0.0:
    # scale = (data[0] + 1) / 500
    scale = 0.0001 if data[0] < 1.0 else data[0] / 500
    data = data + np.random.normal(scale=scale, size=data.shape)

    dens = gaussian_kde(data)
    xs = np.linspace(data.min(), data.max(), 400)
    dens.covariance_factor = lambda: smoothing
    try:
        dens._compute_covariance()
    except:  # noqa: E722
        # There probably wasn't enough data/enough diversity in the data to
        # accurately do this
        pass
    return xs, dens(xs)


def plot_refs(  # noqa: C901
    refs: list[
        reno.components.Reference
        | list[reno.components.Reference]
        | tuple[reno.components.Reference]
    ],
    cols: int = None,
    rows: int = None,
    legends: bool = True,
    **figargs,
) -> plt.Figure:
    """Render a bunch of subplots containing the line plots for the values of the passed
    references.

    `refs` can contain both references directly as well as a nested list/tuple of
    references - anything in a nested list is plotted within the same subplot.

    Note that this is distinct from plot_trace_refs and plots only data from the
    last simulation run for the model

    Example:
        >>> plot_refs([m.stock_1, m.stock_2, m.stock_3])

        >>> plot_refs([m.stock_1, (m.stock_2, m.flow_1, m.flow_2), m.stock_3])

    """
    # TODO: would probably be worth making this also take a traces arg (and
    # simply default to last-sim data in model if None)
    multiple_models = False

    names = []
    for ref in refs:
        if isinstance(ref, (tuple, list)):
            names.extend([sub_ref.model.name for sub_ref in ref])
        else:
            names.append(ref.model.name)
    if len(list(set(names))) > 1:
        multiple_models = True

    # if len(list(set([ref.model.name for ref in refs]))) > 1:

    # establish correct grid size/necessary num of cols/rows if not provided
    if len(refs) > 1 and cols is None:
        cols = 2
    elif cols is None:
        cols = 1
    if rows is None:
        rows = math.ceil(len(refs) / cols)

    plt.ioff()
    fig, axs = plt.subplots(rows, cols, **figargs)
    plt.ion()

    if rows == 1 and cols == 1:
        axs = [[axs]]
    elif rows == 1:
        axs = [axs]

    row_i = 0
    col_i = 0

    for ref in refs:
        if not isinstance(ref, (tuple, list)):
            for row in range(ref.value.shape[0]):
                axs[row_i][col_i].plot(ref.value[row])
            if multiple_models:
                axs[row_i][col_i].set_title(f"{ref.model.name}: {ref.name}")
            else:
                axs[row_i][col_i].set_title(ref.name)
        else:
            # allow passing in tuples of refs to plot on same axis
            for sub_ref in ref:
                for row in range(sub_ref.value.shape[0]):
                    if multiple_models:
                        axs[row_i][col_i].plot(
                            sub_ref.value[row],
                            label=f"{sub_ref.model.name}: {sub_ref.name}",
                        )
                    else:
                        axs[row_i][col_i].plot(sub_ref.value[row], label=sub_ref.name)
            if legends:
                axs[row_i][col_i].legend()
            if multiple_models:
                axs[row_i][col_i].set_title(
                    " vs ".join(
                        [f"{sub_ref.model.name}: {sub_ref.name}" for sub_ref in ref]
                    )
                )
            else:
                axs[row_i][col_i].set_title(
                    " vs ".join([sub_ref.name for sub_ref in ref])
                )

        col_i += 1
        if col_i >= cols:
            col_i = 0
            row_i += 1

    fig.tight_layout()
    return fig


class ModelLatex:
    """An interactive widget that displays all of the latex equations of a model.

    Use with `debug=True` to display all of the values in the equation at a particular
    timestep and sample.

    NOTE: start_name is exclusive, stop_name is inclusive.

    Args:
        model (Model): The system dynamics model to get the equations for.
        show_docs (bool): Whether to display docstrings below each relevant equation.
        start_name (str): Don't display any equations before the one specified (in order
            of their definition.) This is mostly unused for now, intended to eventually
            allow displaying sets of graphs "between" equations.
        stop_name (str): Don't display any equations after the one specified (in order
            of their definition.) This is mostly unused for now, intended to eventually
            allow displaying sets of graphs "between" equations.
        t (int): Only used when `debug` is `True`, refers to the timestep to use for
            displaying the current values within each equation.
        sample (int): Only used when `debug` is `True`, refers to which sample to use
            for displaying the current values within each equation.
        debug (bool): If true, display the value of every reference at the specified `t`
            and `sample`.

    Example:
        >>> ModelLatex(my_model).widget

        >>> ModelLatex(my_model, t=3, sample=0, debug=True).widget
    """

    def __init__(
        self,
        model,
        show_docs: bool = False,
        start_name: str = None,
        stop_name: str = None,
        t: int = None,
        sample: int = 0,
        debug: bool = False,
    ):
        self.model = model
        self.widget = InteractiveLatex()
        """This is the actual displayable object, use this to "view" the equations."""

        self.show_docs = show_docs
        self.debug = debug
        self.t = t
        self.sample = sample

        self.start_name = start_name
        self.stop_name = stop_name

        # NOTE: exclusive start, inclusive stop

        self.widget.latex.data = self.latex_data_with_highlight()
        self.widget.refresh_display()
        self.widget.on_row_clicked(self._handle_ilatex_click)

        self._name_clicked_callbacks: list[Callable[[str], None]] = []

        self.handle_refresh = True

    def on_name_clicked(self, callback: Callable[[str], None]):
        """Register an event handler for when an equation is clicked.

        Callbacks should take a single parameter which is the string name of
        the equation/ref that was clicked.
        """
        self._name_clicked_callbacks.append(callback)

    def fire_on_name_clicked(self, name: str):
        """Trigger the event to notify that a reference/eq was clicked on."""
        for callback in self._name_clicked_callbacks:
            callback(name)

    def _handle_ilatex_click(self, i: int):
        hl_name = self.find_equation_name_from_index(i)
        self.fire_on_name_clicked(hl_name)
        if self.handle_refresh:
            new_latex_data = self.latex_data_with_highlight(hl_name)
            self.widget.latex.data = new_latex_data
            self.widget.refresh_display()

    def find_equation_name_from_index(self, i: int) -> str:
        """Get the name of the reference/equation for the clicked index.

        Returns ``None`` if not found or outside the start/stop bounds."""
        # NOTE: exclusive start, inclusive stop

        names = self._equation_lines_refname_reference()
        if i >= len(names):
            return None
        return names[i]

    def _equation_lines_refname_reference(self) -> list[str]:
        """Create a list of reference names that correspond to the equation lines
        in the latex. This is influenced based on the presence of docstrings (and
        whether show_docs is enabled) etc.

        The number of lines that are related to a particular reference should align
        with and equal the number of times that that reference name appears in the
        returned list.

        For example, a flow with a docstring and show_docs=True will add that flow's
        name twice.

        NOTE: start_name is exclusive, stop_name is inclusive.
        """
        names = []
        started = True if self.start_name is None else False
        stopped = False

        for ref in (
            self.model.all_vars()
            + self.model.all_flows()
            + self.model.all_stocks()
            + self.model.metrics
        ):
            if hasattr(ref, "implicit") and ref.implicit:
                continue

            # logic to only process within specified "name range"
            if not started:
                if ref.qual_name() == self.start_name:
                    started = True
                continue  # exclusive start (don't begin until next)
            if stopped:
                continue
            else:
                if ref.qual_name() == self.stop_name:
                    stopped = True  # inclusive stop (won't stop until next)

            if isinstance(ref, reno.components.Stock):
                # add a line for each equation in the stock
                if not self.debug:
                    for equation in ref.equations():
                        names.append(ref.qual_name())
                else:
                    names.append(ref.qual_name())
            else:
                names.append(ref.qual_name())
            if ref.doc is not None and self.show_docs:
                names.append(ref.qual_name())
        return names

    def latex_data_with_highlight(self, hl_name: str = None) -> str:
        """Construct the latex string with highlighting for the specified name."""
        kwargs = {}
        if hl_name is not None:
            kwargs["hl"] = hl_name
        if self.debug:
            kwargs["t"] = self.t
            kwargs["sample"] = self.sample

        string = "$\n\\begin{align*}\n"

        for ref in (
            self.model.all_vars()
            + self.model.all_flows()
            + self.model.all_stocks()
            + self.model.metrics
        ):
            if hasattr(ref, "implicit") and ref.implicit:
                continue

            # quick cheat to take start/stop names into account by ensuring
            # they're in the aligned list of names we're taking care of
            if ref.qual_name() not in self._equation_lines_refname_reference():
                continue

            highlight = ref.qual_name() == hl_name
            if isinstance(ref, reno.components.Stock):
                # stocks have to be handled differently because they have
                # multiple equations
                if self.debug:
                    equation_str = ref.debug_equation(**kwargs)
                    string += latex_eqline_wrap(equation_str, highlight)
                else:
                    for equation in ref.equations(**kwargs):
                        string += latex_eqline_wrap(equation, highlight)

                if ref.doc is not None and self.show_docs:
                    string += latex_eqline_wrap_doc(
                        ref.name.replace("_", "\\_")
                        + ": "
                        + ref.doc.replace("_", "\\_"),
                        highlight,
                    )
            else:
                if self.debug:
                    equation_str = ref.debug_equation(**kwargs)
                else:
                    equation_str = ref.equation(**kwargs)

                string += latex_eqline_wrap(equation_str, highlight)
                if ref.doc is not None and self.show_docs:
                    string += latex_eqline_wrap_doc(
                        ref.doc.replace("_", "\\_"), highlight
                    )

        string += "\\end{align*}$"
        return string


# class DebugModelLatex:
#     # NOTE: not really ready yet, intent is to make it so when you click on an
#     # equation, it displays all relevant reference graphs "below" that
#     # particular equation. This involves displaying separate model latex objects
#     # above and below with appropriately set "start" and "stop" names.
#     def __init__(self, model):
#         self.model = model
#         self.set1 = ModelLatex(model)
#         self.set2 = ModelLatex(
#             model, start_name=model.stocks[-1], stop_name=model.stocks[-1]
#         )
#
#         self.set1.handle_refresh = False
#         self.set2.handle_refresh = False
#         self.set1.on_name_clicked(self._on_name_clicked)
#         self.set2.on_name_clicked(self._on_name_clicked)
#
#         self.selected_name = None
#
#         self.inner = ipw.Output()
#         self.widget = ipw.Output()
#         # self.widget = ipw.VBox([self.set1.widget, self.inner])#, self.set2.widget])
#
#     def refresh(self):
#         with self.widget:
#             clear_output(True)
#             display(self.set1.widget)
#             # display(self.set1.widget)
#
#     def _on_name_clicked(self, name: str):
#         print("Yo!", name)
#         self.selected_name = name
#         self.set1.stop_name = name
#         self.set2.start_name = name
#
#         set1_data = self.set1.latex_data_with_highlight(name)
#         set2_data = self.set2.latex_data_with_highlight(name)
#
#         self.set1.widget.latex_data = set1_data
#         self.set2.widget.latex_data = set2_data
#         self.set1.widget.refresh_display()
#         self.set2.widget.refresh_display()
#
#         self.refresh()
#
#     def local_plots_for_ref(self, ref, size: int = 2):
#         # size is distance in each direction that we collect for
#
#         # get all relevant refs
#         refs = [ref]
#         refs.extend(ref.seek_refs())
#
#         fig, axs = plt.subplots(1, len(refs))
#
#         domain = [
#             max(self.t - size, 0),
#             min(self.t + size, len(refs[0].value.shape[1])),
#         ]
#
#         for i, l_ref in enumerate(refs):
#             axs[0][i].plot(l_ref.value[self.sample][domain])
#             axs[0][i].set_title(l_ref.name)


@dataclass
class ReferenceEditor:
    """A textbox associated with a reference, used for modifying an equation
    in a visual interface.

    Primarily only used in explorer now.
    """

    model: "reno.model.Model"
    ref_name: str
    is_init: bool
    control: ipw.Text | pn.widgets.TextInput = None

    def get_ref(self) -> reno.components.TrackedReference:
        """Get the underling TrackedReference object associated with this editor."""
        if not self.is_init:
            return getattr(self.model, self.ref_name)
        else:
            return getattr(self.model, self.ref_name.removesuffix("_0"))

    def get_eq(self) -> reno.components.EquationPart:
        """Get the equation from the TrackedReference, (takes into account if
        this is describing the init or the eq itself)."""
        ref = self.get_ref()
        if not self.is_init:
            return ref.eq
        else:
            return ref.init

    def get_eq_str(self) -> str:
        """Get the string version of the current equation."""
        eq = self.get_eq()
        if eq is None:
            return ""
        if isinstance(eq, Scalar):
            return str(eq.value)
        return str(eq)

    def parse_str_to_eq(self) -> reno.components.EquationPart:
        """Get the equivalent EquationPart equation for the string in the control."""
        # result = parse_class_or_scalar(self.control.value)
        result = parse(self.control.value)

        if isinstance(result, (float, int)):
            result = Scalar(result)
        return result

    def assign_value_from_control(self):
        """Set the underlying reference's equation based on the current string
        in the control."""
        result = self.parse_str_to_eq()

        if not self.is_init:
            self.get_ref().eq = result
        else:
            self.get_ref().init = result


class ModelViewer:
    """An interface for viewing a model, modifying any free variables, and simulating live.

    Leaving for now, recommend using Explorer/explorer widgets instead
    """

    def __init__(
        self,
        model,
        exclude_vars: list[str] = None,
        plot_refs: list[
            reno.components.Reference
            | list[reno.components.Reference]
            | tuple[reno.components.Reference]
        ] = None,
    ):
        if exclude_vars is None:
            exclude_vars = []
        self.model = model
        self.exclude_vars = exclude_vars

        self.graphviz_out = ipw.Output()
        self.latex_out = ModelLatex(model)

        self.plots_out = ipw.Output()
        self.metrics_out = ipw.Output()

        self.reference_editors = []
        self.vars_controls = ipw.VBox()
        self.create_variable_controls()

        self.simulate_button = ipw.Button(description="Simulate")
        self.n = ipw.IntText(value=model.n, layout=Layout(width="50px"))
        self.steps = ipw.IntText(value=model.steps, layout=Layout(width="70px"))

        self.show_vars = ipw.Checkbox(description="Show variables", value=False)
        self.show_vars.observe(lambda x: self.rerender_graph(), names=["value"])

        self.plot_refs = plot_refs

        self.widget = ipw.VBox(
            [
                self.show_vars,
                ipw.HBox([self.graphviz_out, self.latex_out.widget]),
                ipw.HBox(
                    [
                        ipw.VBox(
                            [
                                self.vars_controls,
                                ipw.HBox([self.simulate_button, self.n, self.steps]),
                            ]
                        ),
                        ipw.HBox(
                            [
                                self.plots_out,
                                self.metrics_out,
                            ]
                        ),
                    ]
                ),
            ]
        )
        self.simulate_button.on_click(self.simulate_button_press)

        self.rerender_graph()
        self.rerender_equations()

    def simulate_button_press(self, widget):
        """Event handler for when the button labeled "simulate" is pressed, not a
        function simulating a button being pressed..."""
        self.assign_controls()
        self.rerun(self.n.value, self.steps.value)

    def rerender_graph(self):
        with self.graphviz_out:
            clear_output(True)
            display(
                self.model.graph(self.show_vars.value, exclude_vars=self.exclude_vars)
            )

    def rerender_equations(self):
        self.latex_out.latex_data_with_highlight()
        # with self.latex_out:
        #     clear_output(True)
        #     display(self.model)

    def rerun(self, n, steps):
        plt.close("all")
        self.rerender_graph()
        self.rerender_equations()

        self.model.simulate(n, steps, quiet=True)

        with self.plots_out:
            clear_output(True)
            if self.plot_refs is None:
                display(self.model.plot_stocks())
            else:
                display(plot_refs(self.plot_refs))

        with self.metrics_out:
            clear_output(True)
            # TODO: better options than print_metrics
            # self.model.print_metrics()

    def assign_controls(self):
        for ref_editor in self.reference_editors:
            ref_editor.assign_value_from_control()

    def create_variable_controls(self):
        """Set up a bunch of ReferenceEditors for the free variables."""
        style = {"description_width": "initial"}
        controls = []
        for ref_name in self.model.free_refs():
            editor = ReferenceEditor(
                self.model, ref_name, self.model._is_init_ref(ref_name)
            )
            # ref = editor.get_ref()
            control = ipw.Text(
                description=ref_name, value=str(editor.get_eq()), style=style
            )
            editor.control = control

            self.reference_editors.append(editor)
            controls.append(control)
        self.vars_controls.children = controls



"""Tests for the math operations."""

import numpy as np
import pytest

import reno
from reno import model, ops
from reno.components import Flow, Metric, Piecewise, Scalar, Stock, TimeRef, Variable


def test_sum_on_matrix():
    """Running ops.sum on a matrix should give you a row-wise sum."""
    v = Variable()
    v.value = np.array([[0, 1, 2], [1, 2, 3]])

    assert (ops.sum(v.timeseries).eval(3) == np.array([[3, 6]])).all()


def test_sum_on_vector():
    """Running ops.sum on a vector (static variable) should give you the row-wise
    "sum" which is just the value times the number of timesteps."""
    v = Variable()
    v.value = np.array([2, 3])

    assert (ops.sum(ops.orient_timeseries(v)).eval(3) == np.array([[8, 12]])).all()


def test_mean_on_vector():
    """Running ops.mean on a vector should give row-wise mean"""
    v = Variable([0, 1, 2, 3])
    assert v.mean().eval() == 1.5


def test_nonzero_on_vector():
    v = Variable([0, 0, 1, 0, 0, 0, 1])
    assert np.array_equal(
        ops.nonzero(v).eval(),
        np.array([2.0, 6.0, np.nan, np.nan, np.nan, np.nan, np.nan]),
        equal_nan=True,
    )


def test_nonzero_on_matrix():
    v = Variable([[0, 0, 1, 0, 0, 0, 1], [1, 0, 0, 1, 0, 0, 1]])
    assert np.array_equal(
        ops.nonzero(v).eval(),
        np.array(
            [
                [2.0, 6.0, np.nan, np.nan, np.nan, np.nan, np.nan],
                [0.0, 3.0, 6.0, np.nan, np.nan, np.nan, np.nan],
            ]
        ),
        equal_nan=True,
    )


def test_sum_on_nonzero_of_matrix():
    """Running a sum on a nonzero result (which will have nan's) should correctly evaluate."""
    v = Variable([[0, 0, 1, 0, 0, 0, 1], [1, 0, 0, 1, 0, 0, 1]])
    assert np.array_equal(
        ops.nonzero(v).sum().eval(),
        np.array([8.0, 9.0]),
    )


def test_nonzero_in_pymc():
    """The nonzero op should run the same in pymc as in normal reno."""
    m = model.Model()
    t = TimeRef()
    with m:
        v = Variable(
            Piecewise(
                [0, 1], [t.not_equal(2) & t.not_equal(6), t.equal(2) | t.equal(6)]
            )
        )
        met = Metric(ops.nonzero(v.timeseries).sum())

    ds1 = m(n=2)
    ds2 = m.pymc(n=2, compute_prior_only=True)

    print(ds2.prior.met.values)
    assert (ds1.met.values == np.array([8.0, 8.0])).all()
    assert (ds1.met.values == ds2.prior.met.values[0]).all()


def test_sum_on_matrix_start_stop():
    """Running ops.sum on a matrix should give you a row-wise sum, correctly
    accounting for specified range."""
    v = Variable()
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert (v.timeseries[2:].sum().eval(5) == np.array([[9, 12]])).all()
    assert (v.timeseries[:2].sum().eval(5) == np.array([[1, 3]])).all()
    assert (v.timeseries[1:3].sum().eval(5) == np.array([[3, 5]])).all()


def test_sum_on_vector_start_stop():
    """Running ops.sum on a vector (static variable) should give you the row-wise
    "sum" which is just the value times the number of timesteps. Correctly
    accounting for specified range."""
    v = Variable()
    v.value = np.array([2, 3])

    assert (v.timeseries[2:].sum().eval(4) == np.array([[6, 9]])).all()
    assert (v.timeseries[:2].sum().eval(4) == np.array([[4, 6]])).all()
    assert (v.timeseries[1:3].sum().eval(4) == np.array([[4, 6]])).all()


def test_static_value_sum():
    """By default, the sum of a static value will just be itself, unless you access
    it as a slice.
    NOTE: changed this behavior by having sum automatically insert slice if static detected.
    NOTE: changed again due to separate timeseries vs vec series concerns, no longer
    makes sense to take a sum of a non vector thing (which you only get for a scalar by
    looking at its full timeseries)
    """
    v = Variable(Scalar(5))

    # assert v.sum().eval(4) == np.array([25])
    assert v.timeseries[:].sum().eval(4) == np.array([25])


def test_sum_of_series_inside_model():
    """The sum of a timeseries inside a model should still work the same as
    outside a model."""
    m = model.Model()
    with m:
        v0 = Variable(2)
        v1 = Variable(v0.timeseries[0:4], dim=4)
        v2 = Variable(v1.sum())
    ds = m()
    assert (ds.v1.values[0][4] == [2, 2, 2, 2]).all()
    assert ds.v2.values[0][4] == 8


def test_series_max():
    """Both API forms should return the correct series max."""
    v = Variable()
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert (v.series_max().eval(4) == ops.series_max(v).eval(4)).all()


def test_slice_on_matrix():
    """Slice bounds should work properly on a matrix."""
    v = Variable()
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert (v.timeseries[:2].eval(5) == np.array([[0, 1], [1, 2]])).all()
    assert v.timeseries[:2].eval(5).shape == np.array([[0, 1], [1, 2]]).shape
    assert (v.timeseries[0:2].eval(5) == np.array([[0, 1], [1, 2]])).all()
    assert v.timeseries[0:2].eval(5).shape == np.array([[0, 1], [1, 2]]).shape
    assert (v.timeseries[3:].eval(5) == np.array([[3, 4], [4, 5]])).all()
    assert v.timeseries[3:].eval(5).shape == np.array([[3, 4], [4, 5]]).shape


def test_slice_on_vector():
    """Slice bounds should work properly on a matrix."""
    v = Variable()
    v.value = np.array([2, 3])

    assert (v.timeseries[:2].eval(3) == np.array([[2, 2], [3, 3]])).all()
    assert v.timeseries[:2].eval(3).shape == np.array([[2, 2], [3, 3]]).shape
    assert (v.timeseries[0:2].eval(3) == np.array([[2, 2], [3, 3]])).all()
    assert v.timeseries[0:2].eval(3).shape == np.array([[2, 2], [3, 3]]).shape
    assert (v.timeseries[3:].eval(5) == np.array([[2, 2, 2], [3, 3, 3]])).all()
    assert v.timeseries[3:].eval(5).shape == np.array([[2, 2, 2], [3, 3, 3]]).shape


def test_slice_on_scalar():
    """Slicing a scalar should result in a 2d array. (Otherwise
    series ops won't work correctly on the expanded slice.)"""
    v = Variable(Scalar(5))
    assert (v.timeseries[:].eval(4) == np.array([[5, 5, 5, 5, 5]])).all()
    assert v.timeseries[:].eval(4).shape == np.array([[5, 5, 5, 5, 5]]).shape


def test_index_directly_on_scalar():
    """Getting an index on a 1d scalar should still work."""
    s = Scalar([1, 2, 3])
    assert s[1].eval() == 2


def test_eq_index_of_timeseries():
    """An index equation needs to be evaluated on a timeseries in order to correctly
    return only one value."""

    m = model.Model()
    with m:
        t = TimeRef()
        f1 = Flow(t + 1)
        f2 = Flow(f1.timeseries[t - 1])

    ds = m()
    assert ds.f2.values.shape == (1, 10)


def test_staggered_index_access():
    """Accessing different columns in each row with index should correctly return one value per row."""
    m = model.Model()
    with m:
        t = TimeRef()
        indexer = Variable(ops.List([1, 2]))
        f1 = Flow(t + indexer)

        met = Metric(f1.timeseries[indexer])

    ds = m(n=2)
    assert (ds.met.values == [2, 4]).all()


def test_slice_directly_on_scalar():
    """Getting indices on a 1d scalar should still work."""
    s = Scalar([1, 2, 3])
    assert (s[1:].eval() == [2, 3]).all()


def test_slice_staticness_static_slice():
    """A slice with defined bounds (or static bounds) of a non-timeseries multidim should be considered
    static."""
    v = Variable([0, 1, 2, 3, 4])

    assert v[1:3].is_static()

    a = Scalar(1)
    b = Scalar(3)

    assert v[a:b].is_static()

    av = Variable(a)
    bv = Variable(b)

    assert v[av:bv].is_static()


def test_slice_staticness_static_slice_timeseries():
    """A slice with defined bounds (or static bounds) of a timeseries should NOT be considered static."""
    v = Variable(None)
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert not v.timeseries[1:3].is_static()

    a = Scalar(1)
    b = Scalar(3)

    assert not v.timeseries[a:b].is_static()

    av = Variable(a)
    bv = Variable(b)

    assert not v.timeseries[av:bv].is_static()


def test_slice_staticness_none_endpoint():
    """A slice should not be considered static if it's endpoint is None or t."""
    v = Variable(None)
    v.value = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 5]])

    assert not v.timeseries[:].is_static()
    assert not v.timeseries[1:].is_static()
    t = TimeRef()
    assert not v.timeseries[:t].is_static()


def test_slice_staticness_of_timeseries_in_model():
    """A timeseries slice within a model should _not_ be considered static."""
    m = model.Model()
    with m:
        v0 = Variable(2)
        v1 = Variable(v0.timeseries[0:4], dim=4)
        v2 = Variable(v1.sum())

    assert m.v0.is_static()
    assert not m.v1.is_static()
    assert not m.v2.is_static()


# TODO: this doesn't work because of ragged results, the only way we could
# support this would be to allow a "serial execution mode" that runs each
# sample independently
@pytest.mark.skip
def test_slice_staggered_lengths():
    """A slice with multiple samples where each sample needs a different length should work?"""
    m = model.Model()
    with m:
        t = TimeRef()
        v1 = Variable(ops.List([2, 4]))
        v2 = Variable(t + 1)
        met = Metric(v2.timeseries[0:v1])

    ds = m(n=2)
    assert (ds.met.values == np.array[[1, 2, np.nan, np.nan], [1, 2, 3, 4]]).all()


def test_sum_of_dist():
    """Sanity check that running a sum on an array from a distribution
    is correctly row-wise and doesn't collapse to single summed value."""
    b = ops.Bernoulli(0.5)
    b.populate(10)
    b.timeseries.sum().eval(5).shape == (10,)


def test_slice_end_t_is_correct_simpler():
    """A slice stop of none should actually technically be t+1 to be inclusive of current timestep? (This test is
    proof of why, getting index 0 shouldn't be nothing)"""
    v = Variable(Scalar(np.array([2, 3])))
    assert (v.timeseries[:].eval(0) == np.array([[2], [3]])).all()
    assert v.timeseries[:].eval(0).shape == np.array([[2], [3]]).shape


def test_slice_end_t_is_correct():
    """A slice stop of none should actually technically be t+1 to be inclusive of current timestep?"""
    t = TimeRef()
    m = model.Model()
    m.v0 = Variable(t + 1)
    m.v1 = Variable(Scalar(5))
    m.m0 = Metric(m.v0[m.v1 :].sum())
    ds = m()
    assert ds.m0.values[0] == 40.0


def test_piecewise_with_int():
    """Piecewise equations should support just directly specifying an integer and
    have it auto-wrapped in a scalar."""
    t = TimeRef()
    m = model.Model()
    m.v0 = Variable(Piecewise([0, 1], [t < 2, t >= 2]))
    m()


def test_interpolation():
    """Interpolation should work in regular reno math."""
    m = model.Model()
    m.v0 = Variable(Scalar([0, 1, 1.5, 2.72, 3.14]))
    m.v1 = Variable(ops.interpolate(m.v0, [1, 2, 3], [3, 2, 0]))
    ds = m(steps=1)
    np.testing.assert_almost_equal(
        ds.v1.values[0], np.array([3.0, 3.0, 2.5, 0.56, 0.0])
    )


def test_interpolation_pymc():
    """Interpolation should work in pymc math."""
    m = model.Model()
    m.v0 = Variable(Scalar([0, 1, 1.5, 2.72, 3.14]))
    m.v1 = Variable(ops.interpolate(m.v0, [1, 2, 3], [3, 2, 0]))
    ds = m.pymc(steps=1, compute_prior_only=True)
    np.testing.assert_almost_equal(
        ds.prior.v1.values[0][0], np.array([3.0, 3.0, 2.5, 0.56, 0.0])
    )


def test_pulse():
    """A pulse should fire at the correct time for the correct amount of time."""
    m = model.Model()
    m.a = Variable(3)
    m.b = Variable(ops.pulse(5, m.a))
    ds = m()
    assert (ds.b.values == [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0]).all()


def test_first_order_delay():
    """The delay1 op should be equivalent to the corresponding manual stock/flow setup
    for a first order delay, and the pymc output should also work."""
    m1 = model.Model()
    m1.delay_time = Variable(5, dtype=float)
    m1.inflow = Flow(ops.pulse(3, 1) * 10, dtype=float)
    m1.delay_stock = Stock()
    m1.outflow = Flow(m1.delay_stock / m1.delay_time)
    m1.delay_stock += m1.inflow
    m1.delay_stock -= m1.outflow
    ds1 = m1(steps=20)

    m2 = model.Model()
    m2.delay_time = Variable(5, dtype=float)
    m2.input = Variable(ops.pulse(3, 1) * 10, dtype=float)
    m2.flow = Flow(ops.delay1(m2.input, m2.delay_time))
    ds2 = m2(steps=20)

    assert (ds1.outflow.values == ds2.flow.values).all()

    ds3 = m2.pymc(steps=20, n=1, compute_prior_only=True)
    np.testing.assert_almost_equal(ds3.prior.flow.values[0], ds2.flow.values)


def test_delays_against_insight_maker():
    """Trying to ensure my delay operation implementations are consistent with how
    insightmaker does them."""

    m = model.Model()
    m.delay_time = Variable(5.0)
    m.hungry = Flow(ops.pulse(3, 1) * 12, dtype=float)
    m.satiated = Flow(ops.delay1(m.hungry, m.delay_time))
    m.satiated3 = Flow(ops.delay3(m.hungry, m.delay_time))

    ds = m(steps=21)

    im_satiated = [
        0,
        0,
        0,
        0,
        2.4,
        1.92,
        1.536,
        1.2288,
        0.98304,
        0.786432,
        0.6291456,
        0.50331648,
        0.402653184,
        0.3221225472,
        0.25769803776,
        0.206158430208,
        0.1649267441664,
        0.13194139533312,
        0.105553116266496,
        0.0844424930131968,
        0.0675539944105575,
    ]

    im_satiated3 = [
        0,
        0,
        0,
        0,
        0,
        0,
        2.592,
        3.1104,
        2.48832,
        1.65888,
        0.995328,
        0.55738368,
        0.297271296,
        0.1528823808,
        0.0764411904,
        0.03737124864,
        0.0179381993472,
        0.00847987605504001,
        0.003957275492352,
        0.001826434842624,
        0.000834941642342401,
    ]

    np.testing.assert_almost_equal(ds.satiated.values.tolist()[0], im_satiated)
    np.testing.assert_almost_equal(ds.satiated3.values.tolist()[0], im_satiated3)


# TODO: add tests for slices and pymc...


def test_multidim_list():
    """Adding a dim to the List distribution should repeat each "row" value to the given number of columns."""
    my_list = ops.List([0, 1, 2])
    my_list.populate(3)
    assert (my_list.value == np.array([0, 1, 2])).all()

    my_list.populate(3, dim=2)
    assert (my_list.value == np.array([[0, 0], [1, 1], [2, 2]])).all()


def test_multidim_categorical():
    """Adding a dim to the Categorical distribution should return a randomized category matrix."""
    my_cats = ops.Categorical([0.25, 0.25, 0.25, 0.25])
    my_cats.populate(4)
    assert my_cats.value.shape == (4,)

    my_cats.populate(4, dim=3)
    assert my_cats.value.shape == (4, 3)


def test_multidim_vector():
    """Specifying a static variable/flow with an array and dims should treat the array as that extra dim."""
    t = TimeRef()
    thing1 = Variable([0, 1, 2, 3, 4], dim=5)
    thing2 = Variable(thing1 + 10 + t, dim=5)
    thing1.populate(1, 2)
    thing2.populate(1, 2)

    thing2.eval(1, save=True)
    assert (thing1.value == [0, 1, 2, 3, 4]).all()
    assert (thing2.value == [[10, 11, 12, 13, 14], [11, 12, 13, 14, 15]]).all()


def test_multidim_vector2():
    """A static variable/flow from a single scalar should still work as above"""
    t = TimeRef()
    thing1 = Variable(4, dim=5)
    thing2 = Variable(thing1 + 10 + t, dim=5)
    thing1.populate(1, 2)
    thing2.populate(1, 2)

    thing2.eval(1, save=True)
    assert (thing1.value == [4, 4, 4, 4, 4]).all()
    assert (thing2.value == [[14, 14, 14, 14, 14], [15, 15, 15, 15, 15]]).all()


def test_normal_population():
    """Populating a normal distribution without specifying dim should correctly populate."""
    norm = ops.Normal(0, 1)
    norm.populate(3)
    assert norm.value.shape == (3,)

    norm.populate(3, dim=2)
    assert norm.value.shape == (3, 2)


def test_basic_normal_pymc():
    """A normal distribution should translate correctly into pymc."""
    m = model.Model()
    m.v0 = Variable(reno.Normal(5, 10))
    m.f0 = Flow(m.v0)
    m.s = Stock()
    m.s += m.f0
    ds = m.pymc(5, compute_prior_only=True)
    ds.prior.v0.values[0].shape == (5,)


def test_normal_not_none():
    """A model with a normal distribution should run corrrectly."""
    m = model.Model()
    m.v1 = Variable(reno.Normal(5, 10))
    m()
    assert m.v1.value[0] is not None


def test_normal_w_dim():
    """A normal distribution on a variable with an extra dimension should populate the
    full matrix as expected."""
    m = model.Model()
    m.v1 = Variable(reno.Normal(10, 5), dim=3)
    m(n=7, steps=4)
    assert m.v1.value.shape == (7, 3)


def test_basic_normal_w_dim_pymc():
    """A normal distribution with a extra dimension should translate correctly into pymc."""
    m = model.Model()
    m.v0 = Variable(reno.Normal(5, 10), dim=3)
    m.f0 = Flow(m.v0)
    m.s = Stock()
    m.s += m.f0
    ds = m.pymc(5, compute_prior_only=True)
    ds.prior.v0.values[0].shape == (5, 3)


def test_normal_w_dim_and_single_sample():
    """A normal distribution on a variable with an extra dimension should populate the
    full matrix as expected, even when num samples is only 1."""
    m = model.Model()
    m.v1 = Variable(reno.Normal(10, 5), dim=3)
    m(n=1, steps=4)
    assert m.v1.value.shape == (1, 3)


def test_normal_w_seq():
    """A normal distribution with per_timestep requested should correctly populate
    a value for each timestep."""
    m = model.Model()
    m.v1 = Variable(reno.Normal(1, 2, per_timestep=True))
    m(n=2, steps=5)
    assert m.v1.value.shape == (2, 5)


def test_normal_w_seq_and_dim():
    """A normal distribution with per_timestep requested as well as an extra dimension
    should correctly populate all the things."""
    m = model.Model()
    m.v1 = Variable(reno.Normal(1, 2, per_timestep=True), dim=4)
    m(n=2, steps=5)
    assert m.v1.value.shape == (2, 5, 4)


def test_normal_w_seq_and_dim_pymc():
    """Running the pymc version of a model with a normal with both timesteps and extra dims
    should work the same way."""
    m = model.Model()
    m.v1 = Variable(reno.Normal(1, 2, per_timestep=True), dim=4)
    ds = m.pymc(n=2, steps=5, compute_prior_only=True)
    assert ds.prior.v1.values[0].shape == (2, 5, 4)


def test_broken_normal_dist_init_example():
    m = model.Model()

    data_dim = 20
    m.mean = Variable(100, dim=data_dim)
    m.sd = Variable(10, dim=data_dim)
    m.v = Variable(reno.Normal(m.mean, m.sd), dim=data_dim)

    m(n=1, steps=4)


def test_normal_w_dim_parameters_works():
    """A normal distribution populating two-valued data dimension with very different values
    should create two different distributions"""
    m = model.Model()
    m.mean = Variable([0, 1000])
    m.sd = Variable(1)
    m.v = Variable(reno.Normal(m.mean, m.sd))

    assert m.v.shape == 2
    ds = m(n=1000, steps=1)
    assert ds.v.values.shape == (1000, 2)
    low_mean = np.mean(ds.v.values[:, 0])
    high_mean = np.mean(ds.v.values[:, 1])

    assert abs(low_mean) < 0.2
    assert abs(high_mean - 1000) < 0.2


# NOTE: I don't think an actual scalar on a computed value is possible, should
# always be a vector
# def test_sum_on_scalar():
#     """Running ops.sum on a scalar (likely from static varible) should give "sum", or the value
#     time the number of timesteps."""
#     v = Variable(Scalar(3))
#     v.value = 3
#     assert (ops.sum(v).eval(3) == 9)


def test_implicit_components_wo_dim():
    """Using extended ops with equations involving multidim shouldn't break (dim
    should somehow extend thru)"""
    m = model.Model()
    m.v0 = Variable(5, dim=3)
    t = TimeRef()
    m.v1 = Variable(m.v0 + t, dim=3)
    m.v2 = Variable(ops.smooth(m.v1, 3))
    m()


def test_implicit_components_wo_dim_pymc():
    """Using extended ops with equations involving multidim shouldn't break (dim
    should somehow extend thru) and the pymc variant should correctly populate
    implicit components even if the normal __call__() hasn't happened yet"""
    m = model.Model()
    m.v0 = Variable(5, dim=3)
    t = TimeRef()
    m.v1 = Variable(m.v0 + t, dim=3)
    m.v2 = Variable(ops.smooth(m.v1, 3))
    m.pymc()


def test_basic_ops_shapes():
    """Operations between components of different shapes should broadcast as expected."""

    v0 = Variable(5, dim=3)
    v1 = Variable(2)
    v2 = Variable(v0 + v1)

    assert (v0 + v1).shape == 3
    assert v2.shape == 3

    assert (v2.eval(0) == [7, 7, 7]).all()


def test_agg_op_shape():
    """An aggregate operation should correctly get the shape down to 1"""
    v0 = Variable(5, dim=3)
    v1 = Variable(v0.sum())

    assert v0.shape == 3
    assert v0.sum().shape == 1
    assert v1.shape == 1
    assert v1.eval(0) == 15


def test_implied_shape_for_scalar():
    """A scalar given a list/array should use that as the shape."""
    assert Scalar([1, 2, 3]).shape == 3

    v = Variable([1, 2, 3])
    assert v.shape == 3


def test_other_agg_op_shapes():
    """The shape after some agg operations should be 1"""
    v0 = Variable([1, 2, 3])
    v1 = Variable(v0.series_min())
    v2 = Variable(v0.series_max())

    assert v1.shape == 1
    assert v1.eval(0) == 1
    assert v2.shape == 1
    assert v2.eval(0) == 3


def test_agg_after_timeseries_shape():
    """Shape of an aggregate on a timeseries should be 1"""
    m = model.Model()
    t = TimeRef()
    m.v0 = Variable(3)
    m.v1 = Variable(m.v0 + t)
    m.post = Metric(m.v1.timeseries.sum())
    assert m.post.eq.shape == 1
    assert m.post.shape == 1
    m(steps=3)
    assert (m.post.value == [[12]]).all()


def test_list_in_op():
    """Using a list distribution variable with math operations should function as expected."""
    m = model.Model()
    t = TimeRef()
    m.v0 = Variable(ops.List([0, 1, 2]))
    m.v1 = Variable(m.v0 + 3 + t)
    ds = m(n=3, steps=1)
    print(ds)
    assert (ds.v1.values == [[3], [4], [5]]).all()


def test_model_w_multidim_list():
    """Initializing values in a model with a list dist with data dim shouldnt' crash."""
    m = model.Model()
    m.v0 = Variable(ops.List([[0, 1], [1, 2], [2, 3]]))
    m(n=3)


def test_multidim_list_in_op():
    """Using a list distribution variable with a datadim with math operations should function
    as expected."""

    m = model.Model()
    t = TimeRef()
    m.v0 = Variable(ops.List([[0, 1], [1, 2], [2, 3]]))
    m.v1 = Variable(m.v0 + 3 + t)

    assert m.v0.shape == 2
    assert m.v1.shape == 2

    ds = m(n=3, steps=1)
    print(m.v0.value)
    print(m.v1.value)
    print(ds)
    assert (ds.v1.values == [[[3, 4]], [[4, 5]], [[5, 6]]]).all()


# TODO: (2025.09.30) tentatively for now I'm okay with this failing - it
# would make it harder to then count free vs independent variables I think
# if you could directly include?
# def test_distribution_inside_op_as_variable_eq():
#     """What happens when you include a distribution as part of an operation directly in
#     equation for a variable?"""
#     m = model.Model()
#     t = TimeRef()
#     m.v0 = Variable(ops.List([[0, 1], [1, 2], [2, 3]]) + t)
#     m()


def test_distribution_with_time_in_eq():
    """What happens when you include a distribution and incorporate a timeref in an
    equation for a variable?"""
    m = model.Model()
    t = TimeRef()
    m.v0 = Variable(ops.List([[0, 1], [1, 2], [2, 3]]))
    m.v1 = Variable(m.v0 - t)
    m(3, 1)


def test_static_astype():
    """astype on a static value should correctly convert the type"""
    m = model.Model()
    m.v0 = reno.Variable(5.5)
    m.v1 = reno.Variable(m.v0.astype(int))
    ds = m()
    assert ds.v1.values[0] == 5


def test_static_astype_pymc():
    """astype on a static value should correctly convert the type in pymc"""
    m = model.Model()
    m.v0 = reno.Variable(5.5)
    m.v1 = reno.Variable(m.v0.astype(int))
    ds = m.pymc(n=1, compute_prior_only=True)
    assert ds.prior.v1.values[0][0] == 5


def test_astype():
    """astype on a non-static value should correctly convert the type"""
    m = model.Model()
    t = TimeRef()
    m.v0 = reno.Variable(5.5 + t)
    m.v1 = reno.Variable(m.v0.astype(int))
    ds = m(steps=3)
    assert (ds.v1.values == [5, 6, 7]).all()


def test_astype_pymc():
    """astype on a static value should correctly convert the type in pymc"""
    m = model.Model()
    t = TimeRef()
    m.v0 = reno.Variable(5.5 + t)
    m.v1 = reno.Variable(m.v0.astype(int))
    ds = m.pymc(n=1, steps=3, compute_prior_only=True)
    assert (ds.prior.v1.values[0] == [5, 6, 7]).all()


def test_stack():
    """Stacking multiple 1dim variables should correctly turn into a multidim value"""
    m = model.Model()
    with m:
        v0, v1, v2 = Variable(1), Variable(2), Variable(3)
        f3 = Flow(ops.stack(v0, v1, v2))

        s0 = Stock()
        f3 >> s0

    assert f3.shape == 3
    ds = m()
    assert (ds.f3.values[0] == [1, 2, 3]).all()
    assert (ds.s0.values[0][2] == [2, 4, 6]).all()


def test_stack_pymc():
    """Stacking multiple 1dim variables should correctly turn into a multidim value in pymc"""
    m = model.Model()
    with m:
        v0, v1, v2 = Variable(1), Variable(2), Variable(3)
        f3 = Flow(ops.stack(v0, v1, v2))

        s0 = Stock()
        f3 >> s0

    assert f3.shape == 3
    ds = m.pymc(1, compute_prior_only=True)
    assert (ds.prior.f3.values[0][0] == [1, 2, 3]).all()
    assert (ds.prior.s0.values[0][0][2] == [2, 4, 6]).all()


def test_stack_n():
    """Stacking statics with a non-sample-static variable in a n>1 case should still work."""
    m = model.Model()
    with m:
        v0 = Variable(ops.Bernoulli(1.0))
        v1 = Variable(2.0)
        v2 = Variable(ops.stack(v0, v1))

    ds1 = m(n=1)
    assert (ds1.v2.values[0] == [1.0, 2.0]).all()

    ds2 = m(n=2)
    assert (ds1.v2.values[0] == [[1.0, 2.0], [1.0, 2.0]]).all()


def test_stack_n_pymc():
    """Stacking statics with a non-sample-static variable in a n>1 case should still work."""
    m = model.Model()
    with m:
        v0 = Variable(ops.Bernoulli(1.0))
        v1 = Variable(2.0)
        v2 = Variable(ops.stack(v0, v1))

    ds1 = m.pymc(n=1, compute_prior_only=True)
    assert (ds1.prior.v2.values[0][0] == [1.0, 2.0]).all()

    ds2 = m(n=2, compute_prior_only=True)
    assert (ds1.prior.v2.values[0][0] == [[1.0, 2.0], [1.0, 2.0]]).all()


def test_math_w_multidim_and_n():
    """Mixing n and multidim for statics shouldn't impact math."""
    m = model.Model()
    with m:
        v0 = Variable([0, 1, 2, 3])
        v1 = Variable(2)
        v2 = Variable(v0 + v1)

    ds1 = m(n=1)
    assert (ds1.v2.values[0] == [2, 3, 4, 5]).all()

    ds2 = m(n=2)
    assert (ds2.v2.values[0] == [2, 3, 4, 5]).all()


def test_math_w_multidim_and_n_w_dynamic():
    m = model.Model()
    with m:
        v0 = Variable(ops.Bernoulli(1.0), dim=3)
        v1 = Variable(2.0)
        v2 = Variable(v0 + v1)

    ds1 = m(n=1)
    assert (ds1.v2.values[0] == [3.0, 3.0, 3.0]).all()

    ds2 = m(n=2)
    assert (ds2.v2.values[0] == [[3.0, 3.0, 3.0], [3.0, 3.0, 3.0]]).all()


def test_math_w_multidim_and_n_w_really_dynamic():
    m = model.Model()
    t = TimeRef()
    with m:
        v0 = Variable(ops.Bernoulli(1.0), dim=3)
        v1 = Variable(1.0 + t)
        v2 = Variable(v0 + v1)

    ds1 = m(n=1)
    assert (ds1.v2.values[0][1] == [3.0, 3.0, 3.0]).all()

    ds2 = m(n=2)
    assert (ds2.v2.values[0][1] == [[3.0, 3.0, 3.0], [3.0, 3.0, 3.0]]).all()


def test_timeseries_slice_nonmetric():
    """Using timeseries in a regular component (rather than a metric computed after simulation)
    should still work."""
    m = model.Model()
    t = TimeRef()
    with m:
        v0 = Variable(t + 2)
        v1 = Variable(v0.timeseries[t - 3 : t - 1].sum())
    ds = m()
    assert (ds.v1.values[0][0:4] == [0, 0, 2, 5]).all()


def test_timeseries_slice_nonmetric_pymc():
    """Using timeseries in a regular component (rather than a metric computed after simulation)
    should still work in pymc."""
    m = model.Model()
    t = TimeRef()
    with m:
        v0 = Variable(t + 2)
        v1 = Variable(v0.timeseries[t - 3 : t - 1].sum())
    ds = m.pymc(1, compute_prior_only=True)
    assert (ds.prior.v1.values[0][0][0:4] == [0, 0, 2, 5]).all()


def test_dynamic_timeseries_static_slice():
    """A timeseries on a regular component with static slice endpoints should correctly compute
    as if the underlying component is indeed dynamic."""
    m = model.Model()
    t = TimeRef()
    with m:
        v0 = Variable(2 + t)
        v1 = Variable(v0.timeseries[0:4], dim=4)
        v2 = Variable(v1.sum())

    ds = m()
    assert (ds.v2.values[0][0:4] == [2, 5, 9, 14]).all()


def test_dynamic_timeseries_static_slice_pymc():
    """A timeseries on a regular component with static slice endpoints should correctly compute
    as if the underlying component is indeed dynamic."""
    m = model.Model()
    t = TimeRef()
    with m:
        v0 = Variable(2 + t)
        v1 = Variable(v0.timeseries[0:4], dim=4)
        v2 = Variable(v1.sum())

    ds = m.pymc(1, compute_prior_only=True)
    assert (ds.prior.v2.values[0][0][0:4] == [2, 5, 9, 14]).all()


def test_static_timeseries_static_slice():
    """A timeseries on a static component with static slice endpoints should correctly compute
    as if the underlying component is actually dynamic (filling up zeros over time)."""
    m = model.Model()
    with m:
        v0 = Variable(2)
        v1 = Variable(v0.timeseries[0:4], dim=4)
        v2 = Variable(v1.sum())

    ds = m()
    assert (ds.v2.values[0][0:4] == [2, 4, 6, 8]).all()


def test_static_timeseries_static_slice_pymc():
    """A timeseries on a static component with static slice endpoints should correctly compute
    as if the underlying component is actually dynamic (filling up zeros over time)."""
    m = model.Model()
    t = TimeRef()
    with m:
        v0 = Variable(2)
        v1 = Variable(v0.timeseries[0:4], dim=4)
        v2 = Variable(v1.sum())

    ds = m.pymc(1, compute_prior_only=True)
    assert (ds.prior.v2.values[0][0][0:4] == [2, 4, 6, 8]).all()



"""Tests for bayesian analysis/SDM -> PYMC compiling/generation"""

# these are needed for pymc_str exec to work
import numpy as np  # noqa: F401
import pymc as pm
import pytensor  # noqa: F401
import pytensor.tensor as pt  # noqa: F401
import pytest  # noqa: F401
from pytensor.ifelse import ifelse  # noqa: F401

from reno import ops
from reno.components import Flow, Metric, Piecewise, Scalar, Stock, TimeRef, Variable
from reno.model import Model


def test_tub_against_base_reno():
    """pymc_model prior samples with basic deterministic tub example should be
    equivalent to the reno version (which in test_sdm shows is equivalent to
    insightmaker)"""

    t = TimeRef()

    m = Model("tub", steps=7)
    m.water = Stock()
    m.faucet = Flow(
        Piecewise([m.water + 5, Scalar(0.0)], [t <= Scalar(4), t > Scalar(4)])
    )
    m.drain = Flow(Scalar(2))

    m.water += m.faucet
    m.water -= m.drain

    with m.pymc_model():  # as pymc_m:
        idata = pm.sample_prior_predictive(1)

    ds = m()
    assert ds.water.values.tolist()[0] == [0.0, 3.0, 9.0, 21.0, 45.0, 93.0, 91.0]
    assert ds.faucet.values.tolist()[0] == [5.0, 8.0, 14.0, 26.0, 50.0, 0.0, 0.0]

    assert (
        idata.prior.water.sel(chain=0).values.tolist()[0] == ds.water.values.tolist()[0]
    )
    assert (
        idata.prior.faucet.sel(chain=0).values.tolist()[0]
        == ds.faucet.values.tolist()[0]
    )


def test_multidim_model_against_base_reno(multidim_model_determ):
    """A model with stuff with vector dimensions should be equivalent to the reno version."""

    ds1 = multidim_model_determ()
    ds2 = multidim_model_determ.pymc(1)

    assert (ds1.s.values == ds2.prior.s.sel(chain=0).values).all()
    assert (ds1.v1.values == ds2.prior.v1.sel(chain=0).values).all()
    assert (ds1.v2.values == ds2.prior.v2.sel(chain=0).values).all()
    assert (ds1.v3.values == ds2.prior.v3.sel(chain=0).values).all()


def test_pymc_str_equiv_to_pymc_model():
    """Evaling and running pymc_str code should be equivalent to pymc_model"""

    t = TimeRef()

    m = Model("tub", steps=7)
    m.water = Stock()
    m.faucet = Flow(
        Piecewise([m.water + 5, Scalar(0.0)], [t <= Scalar(4), t > Scalar(4)])
    )
    m.drain = Flow(Scalar(2))

    m.water += m.faucet
    m.water -= m.drain

    vars = globals()
    exec(m.pymc_str(), vars)
    with vars["tub_pymc_m"]:
        idata1 = pm.sample_prior_predictive(1)

    with m.pymc_model():
        idata2 = pm.sample_prior_predictive(1)

    assert (
        idata1.prior.water.sel(chain=0).values == idata2.prior.water.sel(chain=0).values
    ).all()
    assert (
        idata1.prior.faucet.sel(chain=0).values
        == idata2.prior.faucet.sel(chain=0).values
    ).all()


def test_bernoulli_udpates():
    """From notebook 28, confirming a boolean based RV can correctly update based on downstream
    likelihoods."""

    m = Model()
    m.decision = Variable(ops.Bernoulli(0.5))
    m.f = Flow(
        Piecewise(
            [Scalar(100), Scalar(0)],
            [ops.eq(m.decision, Scalar(1)), ops.eq(m.decision, Scalar(0))],
        )
    )
    m.s = Stock()
    m.s += m.f

    with m.pymc_model() as m1:
        pm.Normal("likelihood", m1["s"][-1], 100, observed=[900])
        prior = pm.sample_prior_predictive(100)
        posterior = pm.sample(100)

    assert (
        prior.prior.decision.values.sum() >= 20
        and prior.prior.decision.values.sum() <= 80
    )
    assert (posterior.posterior.decision.values == 1).all()


def test_bernoulli_updates_with_obs_op():
    """Similar to test_bernoulli_updates, but directly passing observations to pymc_model."""

    m = Model()
    m.decision = Variable(ops.Bernoulli(0.5))
    m.f = Flow(
        Piecewise(
            [Scalar(100), Scalar(0)],
            [ops.eq(m.decision, Scalar(1)), ops.eq(m.decision, Scalar(0))],
        )
    )
    m.s = Stock()
    m.s += m.f

    m.final_s = Metric(ops.index(m.s, Scalar(-1)))

    with m.pymc_model([ops.Observation(m.final_s, 100, [900])]):
        posterior = pm.sample(100)
    assert (posterior.posterior.decision.values == 1).all()


def test_bernoulli_updates_faster_compile():
    """Similar to test_bernoulli_updates_with_obs_op, but ensuring 'lesser compile optimization'
    doesn't break."""

    m = Model()
    m.decision = Variable(ops.Bernoulli(0.5))
    m.f = Flow(
        Piecewise(
            [Scalar(100), Scalar(0)],
            [ops.eq(m.decision, Scalar(1)), ops.eq(m.decision, Scalar(0))],
        )
    )
    m.s = Stock()
    m.s += m.f

    m.final_s = Metric(ops.index(m.s, Scalar(-1)))

    posterior = m.pymc(
        100, compile_faster=True, observations=[ops.Observation(m.final_s, 100, [900])]
    )
    assert (posterior.posterior.decision.values == 1).all()


# TODO: all the testing models from notebook 23 to ensure ordering


def test_pymc_multi_submodels():
    """Evaling and prior sampling from a model with several submodels shouldn't fail."""
    a = Model("a")
    a.s1 = Stock()
    a.f1, a.f2 = Flow(Scalar(3)), Flow(Scalar(2))
    a.s1 += a.f1
    a.s1 -= a.f2
    a.s2 = Stock()

    b = Model("b")
    a.b = b
    b.s1_b = Stock()
    b.f1_b, b.f2_b = Flow(), Flow(Scalar(1))
    b.f1_b.eq = a.f2
    b.s1_b += b.f1_b
    b.s1_b -= b.f2_b

    c = Model("c")
    a.c = c
    c.s1_c = Stock()
    c.f1_c, c.f2_c = Flow(), Flow(Scalar(1))
    c.f1_c.eq = a.f2
    c.s1_c += c.f1_c
    c.s1_c -= c.f2_c

    a.s2 += b.f2_b
    a.s2 += c.f2_c

    vars = globals()
    exec(a.pymc_str(), vars)
    with vars["a_pymc_m"]:
        idata1 = pm.sample_prior_predictive(1)

    with a.pymc_model():
        idata2 = pm.sample_prior_predictive(1)

    assert (
        idata1.prior.s2.sel(chain=0).values == idata2.prior.s2.sel(chain=0).values
    ).all()


def test_pymc_str_historical_refs():
    """PyMC string code for a model referencing historical values should
    correctly run."""

    t = TimeRef()

    m = Model()
    m.v0 = Variable(Scalar(1))
    m.v1 = Variable(t)
    m.v2 = Variable(m.v1.history(t - 1))
    m.v3 = Variable(m.v1.history(t - 2))
    m.v4 = Variable(m.v2.history(t - m.v0))

    ds = m()

    vars = globals()
    exec(m.pymc_str(), vars)
    with vars["pymc_m"]:
        idata = pm.sample_prior_predictive(1)

    assert idata.prior.v0.sel(chain=0).values.tolist() == ds.v0.values.tolist()
    assert idata.prior.v1.sel(chain=0).values.tolist() == ds.v1.values.tolist()
    assert idata.prior.v2.sel(chain=0).values.tolist() == ds.v2.values.tolist()
    assert idata.prior.v3.sel(chain=0).values.tolist() == ds.v3.values.tolist()
    assert idata.prior.v4.sel(chain=0).values.tolist() == ds.v4.values.tolist()


def test_pymc_model_historical_refs():
    """The pymc converted model should run equivalently to the pymc str code. (and
    by transitive property should be equivalent to the normal reno model)"""

    t = TimeRef()

    m = Model()
    m.v0 = Variable(Scalar(1))
    m.v1 = Variable(t)
    m.v2 = Variable(m.v1.history(t - 1))
    m.v3 = Variable(m.v1.history(t - 2))
    m.v4 = Variable(m.v2.history(t - m.v0))

    vars = globals()
    exec(m.pymc_str(), vars)
    with vars["pymc_m"]:
        idata1 = pm.sample_prior_predictive(1)

    with m.pymc_model():
        idata2 = pm.sample_prior_predictive(1)

    assert (
        idata1.prior.v0.sel(chain=0).values == idata2.prior.v0.sel(chain=0).values
    ).all()
    assert (
        idata1.prior.v1.sel(chain=0).values == idata2.prior.v1.sel(chain=0).values
    ).all()
    assert (
        idata1.prior.v2.sel(chain=0).values == idata2.prior.v2.sel(chain=0).values
    ).all()
    assert (
        idata1.prior.v3.sel(chain=0).values == idata2.prior.v3.sel(chain=0).values
    ).all()
    assert (
        idata1.prior.v4.sel(chain=0).values == idata2.prior.v4.sel(chain=0).values
    ).all()


def test_pymc_multi_submodels_with_metrics(tub_multimodel):
    """Make sure submodel metrics correctly get handled in pymc."""
    run = tub_multimodel.pymc(n=1, compute_prior_only=True)
    assert "tub_final_water_level" in run.prior.keys()


def test_pymc_slice_in_metric():
    """Postmeasurements using slices should work the same in pymc as in reno math."""
    t = TimeRef()
    m = Model()
    m.v0 = Variable(t + 1)
    m.m0 = Metric(m.v0[-4:-1].sum())

    run_reno = m()
    run_pymc = m.pymc()
    assert (run_pymc.prior.v0.sel(chain=0).values == run_reno.v0.values).all()
    assert (run_pymc.prior.m0.sel(chain=0).values == run_reno.m0.values).all()


def test_pymc_slice_in_metric_no_upper_bound():
    """Postmeasurements using slices should work the same in pymc as in reno math."""
    t = TimeRef()
    m = Model()
    m.v0 = Variable(t + 1)
    m.m0 = Metric(m.v0[-4:].sum())

    run_reno = m()
    run_pymc = m.pymc()
    assert (run_pymc.prior.v0.sel(chain=0).values == run_reno.v0.values).all()
    assert (run_pymc.prior.m0.sel(chain=0).values == run_reno.m0.values).all()


def test_pymc_slice_with_vars_in_metric():
    """Postmeasurements using slices should work the same in pymc as in reno math."""
    t = TimeRef()
    m = Model()
    m.v0 = Variable(t + 1)
    m.v1 = Variable(Scalar(5))
    m.m0 = Metric(m.v0.timeseries[m.v1 :].sum())

    run_reno = m()
    run_pymc = m.pymc()
    assert (run_pymc.prior.v0.sel(chain=0).values == run_reno.v0.values).all()
    assert (run_pymc.prior.m0.sel(chain=0).values == run_reno.m0.values).all()


def test_pymc_historical_of_multidim():
    """The init should correctly initialize multidim zero vectors for a historical val
    of a multidim variable"""
    t = TimeRef()
    m = Model()
    m.v0 = Variable([2, 3])
    m.v1 = Variable(t * m.v0)
    m.v2 = Variable(m.v1.history(t - 1))
    m.s = Stock()
    m.s += m.v2

    assert m.v0.shape == 2
    assert m.v1.shape == 2
    assert m.v2.shape == 2
    assert m.s.shape == 2

    run = m.pymc(compute_prior_only=True)
    assert (run.prior.s.values[0][0][-1] == [56, 84]).all()


def test_pymc_historical_of_multidim_w_taps():
    """The init should correctly initialize multidim zero vectors for a historical val
    of a multidim variable. (This one actually uses taps)"""
    t = TimeRef()
    m = Model()
    m.v0 = Variable([2, 3])
    m.v1 = Variable(t * m.v0)
    m.v2 = Variable(m.v1.history(t - 2))
    m.s = Stock()
    m.s += m.v2

    assert m.v0.shape == 2
    assert m.v1.shape == 2
    assert m.v2.shape == 2
    assert m.s.shape == 2

    run = m.pymc(compute_prior_only=True)
    assert (run.prior.s.values[0][0][-1] == [42, 63]).all()


def test_implicit_metrics():
    """Specifying an observation with an equation rather than a metric should create an implicit metric."""
    t = TimeRef()
    m = Model()
    with m:
        v0 = Variable(ops.Normal(5, 2))
        v1 = Variable(t + v0)

    ds = m.pymc(1000, observations=[ops.Observation(m.v1.timeseries[3], 1, [5.0])])

    assert "observation_0" in ds.prior
    assert "observation_0" not in [metric.name for metric in m.metrics]


def test_implicit_timeref():
    """A model that has a time ref added because of an operation should correctly pick that up in pymc land"""
    m = Model()
    with m:
        v0 = Variable(ops.repeated_pulse(0, 3))

    print(m.find_timeref_name())
    ds = m.pymc(compute_prior_only=True)
